# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "bewu@mozilla.com",
    "start_date": datetime.datetime(2020, 10, 31, 0, 0),
    "end_date": None,
    "email": ["bewu@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_google_analytics_derived",
    default_args=default_args,
    schedule_interval="0 23 * * *",
) as dag:

    ga_derived__blogs_daily_summary__v1 = bigquery_etl_query(
        task_id="ga_derived__blogs_daily_summary__v1",
        destination_table="blogs_daily_summary_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__blogs_empty_check__v1 = bigquery_etl_query(
        task_id="ga_derived__blogs_empty_check__v1",
        destination_table=None,
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        parameters=["submission_date:DATE:{{ds}}"],
        sql_file_path="sql/moz-fx-data-marketing-prod/ga_derived/blogs_empty_check_v1/query.sql",
        dag=dag,
    )

    ga_derived__blogs_goals__v1 = bigquery_etl_query(
        task_id="ga_derived__blogs_goals__v1",
        destination_table="blogs_goals_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__blogs_landing_page_summary__v1 = bigquery_etl_query(
        task_id="ga_derived__blogs_landing_page_summary__v1",
        destination_table="blogs_landing_page_summary_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__blogs_sessions__v1 = bigquery_etl_query(
        task_id="ga_derived__blogs_sessions__v1",
        destination_table="blogs_sessions_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__www_site_downloads__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_downloads__v1",
        destination_table="www_site_downloads_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__www_site_empty_check__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_empty_check__v1",
        destination_table=None,
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        parameters=["submission_date:DATE:{{ds}}"],
        sql_file_path="sql/moz-fx-data-marketing-prod/ga_derived/www_site_empty_check_v1/query.sql",
        dag=dag,
    )

    ga_derived__www_site_events_metrics__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_events_metrics__v1",
        destination_table="www_site_events_metrics_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__www_site_hits__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_hits__v1",
        destination_table="www_site_hits_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__www_site_landing_page_metrics__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_landing_page_metrics__v1",
        destination_table="www_site_landing_page_metrics_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__www_site_metrics_summary__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_metrics_summary__v1",
        destination_table="www_site_metrics_summary_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__www_site_page_metrics__v1 = bigquery_etl_query(
        task_id="ga_derived__www_site_page_metrics__v1",
        destination_table="www_site_page_metrics_v1",
        dataset_id="ga_derived",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    ga_derived__blogs_daily_summary__v1.set_upstream(ga_derived__blogs_goals__v1)

    ga_derived__blogs_daily_summary__v1.set_upstream(ga_derived__blogs_sessions__v1)

    ga_derived__blogs_goals__v1.set_upstream(ga_derived__blogs_empty_check__v1)

    ga_derived__blogs_landing_page_summary__v1.set_upstream(ga_derived__blogs_goals__v1)

    ga_derived__blogs_landing_page_summary__v1.set_upstream(
        ga_derived__blogs_sessions__v1
    )

    ga_derived__blogs_sessions__v1.set_upstream(ga_derived__blogs_empty_check__v1)

    ga_derived__www_site_downloads__v1.set_upstream(ga_derived__www_site_hits__v1)

    ga_derived__www_site_events_metrics__v1.set_upstream(ga_derived__www_site_hits__v1)

    ga_derived__www_site_hits__v1.set_upstream(ga_derived__www_site_empty_check__v1)

    ga_derived__www_site_landing_page_metrics__v1.set_upstream(
        ga_derived__www_site_downloads__v1
    )

    ga_derived__www_site_landing_page_metrics__v1.set_upstream(
        ga_derived__www_site_hits__v1
    )

    ga_derived__www_site_metrics_summary__v1.set_upstream(
        ga_derived__www_site_downloads__v1
    )

    ga_derived__www_site_page_metrics__v1.set_upstream(ga_derived__www_site_hits__v1)



import pytest
from click.testing import CliRunner

from bigquery_etl.cli.view import view


class TestPublish:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_publish_invalid_view(self, runner, tmp_path):
        (tmp_path / "view.sql").write_text("SELECT 42 as test")
        result = runner.invoke(view, ["publish", tmp_path.as_posix(), "--dry-run"])
        assert result.exit_code == 1
        assert "does not appear to be a CREATE OR REPLACE VIEW" in result.output

    def test_publish_valid_view(self, runner, tmp_path):
        # In order to be agnostic with respect to individual projects in GCP,
        # we'll try to dry-run a query with a resource that certainly should not
        # exist.
        (tmp_path / "view.sql").write_text(
            """
            CREATE OR REPLACE VIEW test.test.test AS
            SELECT 42 as test
        """
        )
        result = runner.invoke(view, ["publish", tmp_path.as_posix(), "--dry-run"])
        assert result.exit_code == 1
        assert "Not found" in result.exc_info[1].message



"""Utility functions used in generating usage queries on top of Glean."""

import logging
import os
import re
from pathlib import Path

from jinja2 import Environment, PackageLoader

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.bigquery_id import sql_table_id  # noqa E402


def render(sql_filename, **kwargs) -> str:
    """Render a given template query using Jinja."""
    env = Environment(loader=PackageLoader("bigquery_etl", "glean_usage/templates"))
    main_sql = env.get_template(sql_filename)
    return reformat(main_sql.render(**kwargs))


def write_sql(output_dir, full_table_id, basename, sql):
    """Write out a query to a location based on the table ID.

    :param output_dir:    Base target directory (probably sql/moz-fx-data-shared-prod/)
    :param full_table_id: Table ID in project.dataset.table form
    :param basename:      The name to give the written file (like query.sql)
    :param sql:           The query content to write out
    """
    d = Path(os.path.join(output_dir, *list(full_table_id.split(".")[-2:])))
    d.mkdir(parents=True, exist_ok=True)
    target = d / basename
    logging.info(f"Writing {target}")
    with target.open("w") as f:
        f.write(sql)
        f.write("\n")


def list_baseline_tables(client, pool, project_id, only_tables, table_filter):
    """Make parallel BQ API calls to grab all Glean baseline stable tables.

    See `util.standard_args` for more context on table filtering.

    :param client:       A BigQuery client object
    :param pool:         A process pool for handling concurrent calls
    :param project_id:   Target project
    :param only_tables:  An iterable of globs in `<stable_dataset>.<table>` format
    :param table_filter: A function for determining whether to process a table
    :return:             A list of matching tables
    """
    if only_tables and not _contains_glob(only_tables):
        # skip list calls when only_tables exists and contains no globs
        return [f"{project_id}.{t}" for t in only_tables if table_filter(t)]
    if only_tables and not _contains_glob(
        _extract_dataset_from_glob(t) for t in only_tables
    ):
        # skip list_datasets call when only_tables exists and datasets contain no globs
        stable_datasets = {
            f"{project_id}.{_extract_dataset_from_glob(t)}" for t in only_tables
        }
    else:
        stable_datasets = [
            d.reference.dataset_id for d in client.list_datasets(project_id)
        ]
    stable_datasets = [d for d in stable_datasets if d.endswith("_stable")]
    return [
        sql_table_id(t)
        for tables in pool.map(client.list_tables, stable_datasets)
        for t in tables
        if table_filter(f"{t.dataset_id}.{t.table_id}")
        and t.table_id == "baseline_v1"
        and t.labels.get("schema_id") == "glean_ping_1"
    ]


def table_names_from_baseline(baseline_table):
    """Return a dict with full table IDs for derived tables and views.

    :param baseline_table: stable table ID in project.dataset.table form
    """
    prefix = re.sub(r"_stable\..+", "", baseline_table)
    return dict(
        baseline_table=baseline_table,
        daily_table=f"{prefix}_derived.baseline_clients_daily_v1",
        last_seen_table=f"{prefix}_derived.baseline_clients_last_seen_v1",
        daily_view=f"{prefix}.baseline_clients_daily",
        last_seen_view=f"{prefix}.baseline_clients_last_seen",
    )


def _contains_glob(patterns):
    return any({"*", "?", "["}.intersection(pattern) for pattern in patterns)


def _extract_dataset_from_glob(pattern):
    # Assumes globs are in <dataset>.<table> form without a project specified.
    return pattern.split(".", 1)[0]



import os
import pytest
from click.testing import CliRunner
import yaml

from bigquery_etl.cli.query import (
    create,
    schedule,
    info,
    _queries_matching_name_pattern,
)


class TestQuery:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_create_invalid_path(self, runner):
        with runner.isolated_filesystem():
            with open("foo.txt", "w") as f:
                f.write("")
            result = runner.invoke(create, ["test.query_v1", "--sql_dir=foo.txt"])
            assert result.exit_code == 2

    def test_create_invalid_query_name(self, runner):
        with runner.isolated_filesystem():
            result = runner.invoke(create, ["invalid_query_name"])
            assert result.exit_code == 2

    def test_create_query(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query"])
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod") == ["test"]
            assert os.listdir("sql/moz-fx-data-shared-prod/test") == ["test_query_v1"]
            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

    def test_create_query_with_version(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query_v4"])
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod/test") == ["test_query_v4"]

    def test_create_derived_query_with_view(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test_derived")
            result = runner.invoke(create, ["test_derived.test_query"])
            assert result.exit_code == 0
            assert "test_derived" in os.listdir("sql/moz-fx-data-shared-prod")
            assert "test" in os.listdir("sql/moz-fx-data-shared-prod")
            assert os.listdir("sql/moz-fx-data-shared-prod/test_derived") == [
                "test_query_v1"
            ]
            assert os.listdir("sql/moz-fx-data-shared-prod/test") == ["test_query"]
            assert os.listdir("sql/moz-fx-data-shared-prod/test/test_query") == [
                "view.sql"
            ]

    def test_create_query_as_derived(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            os.mkdir("sql/moz-fx-data-shared-prod/test_derived")
            os.mkdir("sql/moz-fx-data-shared-prod/test")
            result = runner.invoke(create, ["test.test_query"])
            assert result.exit_code == 0
            assert "test_derived" in os.listdir("sql/moz-fx-data-shared-prod")
            assert "test" in os.listdir("sql/moz-fx-data-shared-prod")
            assert os.listdir("sql/moz-fx-data-shared-prod/test_derived") == [
                "test_query_v1"
            ]
            assert os.listdir("sql/moz-fx-data-shared-prod/test") == ["test_query"]
            assert os.listdir("sql/moz-fx-data-shared-prod/test/test_query") == [
                "view.sql"
            ]

    def test_create_query_with_init(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query", "--init"])
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod/test") == ["test_query_v1"]
            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            assert "init.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

    def test_schedule_invalid_path(self, runner):
        with runner.isolated_filesystem():
            result = runner.invoke(schedule, ["/test/query_v1"])
            assert result.exit_code == 2

    def test_schedule_query_non_existing_dag(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/query_v1")
            open("sql/moz-fx-data-shared-prod/test/query_v1/query.sql", "a").close()
            result = runner.invoke(schedule, ["test.query_v1", "--dag=foo"])
            assert result.exit_code == 1

    def test_schedule_query(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            os.mkdir("dags")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            dag_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.com",
                        "start_date": "2020-03-29",
                        "email": ["test@example.org"],
                        "retries": 1,
                    },
                }
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dag_conf))

            result = runner.invoke(
                schedule, ["telemetry_derived.query_v1", "--dag=bqetl_test"]
            )

            assert result.exit_code == 0

    def test_reschedule_query(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            os.mkdir("dags")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            dag_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.com",
                        "start_date": "2020-03-29",
                        "email": ["test@example.org"],
                        "retries": 1,
                    },
                }
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dag_conf))

            result = runner.invoke(schedule, ["telemetry_derived.query_v1"])

            assert result.exit_code == 0

    def test_query_info(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            result = runner.invoke(info, ["*.telemetry_derived.query_v1"])
            assert result.exit_code == 0
            assert "No metadata" in result.output
            assert "path:" in result.output

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            result = runner.invoke(info, ["*.telemetry_derived.query_v1"])
            assert result.exit_code == 0
            assert "No metadata" not in result.output
            assert "description" in result.output
            assert "dag_name: bqetl_test" in result.output

    def test_info_name_pattern(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.mkdir("sql/moz-fx-data-shared-prod/telemetry_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.makedirs("sql/moz-fx-data-shared-prod/foo_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/foo_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            result = runner.invoke(info, ["*.query_*"])
            assert result.exit_code == 0
            assert "foo_derived.query_v2" in result.output
            assert "telemetry_derived.query_v2" in result.output
            assert "telemetry_derived.query_v1" in result.output

            result = runner.invoke(
                info, ["foo_derived.*", "--project-id=moz-fx-data-shared-prod"]
            )
            assert result.exit_code == 0
            assert "foo_derived.query_v2" in result.output
            assert "telemetry_derived.query_v2" not in result.output
            assert "telemetry_derived.query_v1" not in result.output

            result = runner.invoke(info, ["*.query_v2"])
            assert result.exit_code == 0
            assert "foo_derived.query_v2" in result.output
            assert "telemetry_derived.query_v2" in result.output
            assert "telemetry_derived.query_v1" not in result.output

    def test_queries_matching_name_pattern(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.mkdir("sql/moz-fx-data-shared-prod/telemetry_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.makedirs("sql/moz-fx-data-shared-prod/foo_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/foo_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.makedirs("sql/moz-fx-data-test-project/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-test-project/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert len(_queries_matching_name_pattern("*", "sql/", None)) == 4
            assert (
                len(
                    _queries_matching_name_pattern(
                        "*.sql", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 0
            )
            assert (
                len(
                    _queries_matching_name_pattern(
                        "test", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 0
            )
            assert (
                len(
                    _queries_matching_name_pattern(
                        "foo_derived", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 0
            )
            assert (
                len(
                    _queries_matching_name_pattern(
                        "foo_derived*", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 1
            )
            assert len(_queries_matching_name_pattern("*query*", "sql/", None)) == 4
            assert (
                len(
                    _queries_matching_name_pattern(
                        "foo_derived.query_v2", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 1
            )

            assert (
                len(
                    _queries_matching_name_pattern(
                        "telemetry_derived.query_v1", "sql/", "moz-fx-data-test-project"
                    )
                )
                == 1
            )

            assert (
                len(
                    _queries_matching_name_pattern(
                        "moz-fx-data-test-project.telemetry_derived.query_v1",
                        "sql/",
                        None,
                    )
                )
                == 1
            )

            assert (
                len(
                    _queries_matching_name_pattern(
                        "moz-fx-data-test-project.telemetry_derived.*", "sql/", None
                    )
                )
                == 1
            )



"""
Dry run query files.

Passes all queries to a Cloud Function that will run the
queries with the dry_run option enabled.

We could provision BigQuery credentials to the CircleCI job to allow it to run
the queries directly, but there is no way to restrict permissions such that
only dry runs can be performed. In order to reduce risk of CI or local users
accidentally running queries during tests and overwriting production data, we
proxy the queries through the dry run service endpoint.
"""
from functools import cached_property
from multiprocessing.pool import ThreadPool
from os.path import basename, dirname
from urllib.request import urlopen, Request
import glob
import json
import sys

SKIP = {
    # Access Denied
    "sql/moz-fx-data-shared-prod/account_ecosystem_derived/ecosystem_client_id_lookup_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/account_ecosystem_derived/desktop_clients_daily_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/account_ecosystem_restricted/ecosystem_client_id_deletion_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/account_ecosystem_derived/fxa_logging_users_daily_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/activity_stream/impression_stats_flat/view.sql",
    "sql/moz-fx-data-shared-prod/activity_stream/tile_id_types/view.sql",
    "sql/moz-fx-data-shared-prod/monitoring/deletion_request_volume_v1/query.sql",
    "sql/moz-fx-data-shared-prod/monitoring/document_sample_nonprod_v1/query.sql",
    "sql/moz-fx-data-shared-prod/monitoring/schema_error_counts_v1/view.sql",
    "sql/moz-fx-data-shared-prod/monitoring/structured_error_counts_v1/view.sql",
    "sql/moz-fx-data-shared-prod/monitoring/telemetry_missing_columns_v1/view.sql",
    "sql/moz-fx-data-shared-prod/monitoring/telemetry_missing_columns_v2/view.sql",
    "sql/moz-fx-data-shared-prod/pocket/pocket_reach_mau/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/buildhub2/view.sql",
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_content_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_auth_bounce_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_auth_events_v1/query.sql",
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_delete_events_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_delete_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_oauth_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_log_auth_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_log_content_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_log_device_command_events_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/addons_daily_v1/query.sql",
    "sql/moz-fx-data-shared-prod/search_derived/search_clients_last_seen_v1/init.sql",
    "sql/moz-fx-data-shared-prod/search_derived/search_clients_last_seen_v1/query.sql",
    "sql/moz-fx-data-shared-prod/search/search_rfm/view.sql",
    "sql/moz-fx-data-shared-prod/search/search_clients_last_seen_v1/view.sql",
    "sql/moz-fx-data-shared-prod/search/search_clients_last_seen/view.sql",
    "sql/moz-fx-data-shared-prod/firefox_accounts/fxa_amplitude_email_clicks/view.sql",
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_amplitude_export_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_amplitude_user_ids_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/firefox_accounts_derived/fxa_amplitude_user_ids_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/regrets_reporter/regrets_reporter_update/view.sql",
    "sql/moz-fx-data-shared-prod/revenue_derived/client_ltv_v1/query.sql",
    "sql/moz-fx-data-shared-prod/shredder_state/progress/view.sql",
    "sql/moz-fx-data-shared-prod/monitoring/telemetry_distinct_docids_v1/query.sql",
    "sql/moz-fx-data-shared-prod/revenue_derived/client_ltv_normalized/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_derived/customers_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_derived/plans_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_derived/products_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_derived/subscriptions_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/charges_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/credit_notes_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/customers_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/disputes_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/invoices_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/payment_intents_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/payouts_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/plans_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/prices_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/products_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/setup_intents_v1/query.sql",
    "sql/moz-fx-data-shared-prod/stripe_external/subscriptions_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_derived/add_device_events_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_derived/login_flows_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_derived/protected_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_derived/users_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_derived/waitlist_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_external/devices_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_external/subscriptions_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_external/users_v1/query.sql",
    "sql/moz-fx-data-shared-prod/mozilla_vpn_external/waitlist_v1/query.sql",
    "sql/moz-fx-data-shared-prod/monitoring/telemetry_missing_columns_v3/query.sql",
    # Already exists (and lacks an "OR REPLACE" clause)
    "sql/moz-fx-data-shared-prod/org_mozilla_firefox_derived/clients_first_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/org_mozilla_firefox_derived/clients_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/org_mozilla_fenix_derived/clients_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/org_mozilla_vrbrowser_derived/clients_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/core_clients_last_seen_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry/fxa_users_last_seen_raw_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/core_clients_first_seen_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/fxa_users_services_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/messaging_system_derived/cfr_users_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/messaging_system_derived/onboarding_users_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/messaging_system_derived/snippets_users_last_seen_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/messaging_system_derived/whats_new_panel_users_last_seen_v1/init.sql",  # noqa E501
    # Reference table not found
    "sql/moz-fx-data-shared-prod/monitoring/structured_detailed_error_counts_v1/view.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/org_mozilla_firefox_derived/migrated_clients_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/org_mozilla_firefox_derived/incline_executive_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/org_mozilla_firefox/migrated_clients/view.sql",
    # No matching signature for function IF
    "sql/moz-fx-data-shared-prod/static/fxa_amplitude_export_users_last_seen/query.sql",
    # Duplicate UDF
    "sql/moz-fx-data-shared-prod/static/fxa_amplitude_export_users_daily/query.sql",
    # Syntax error
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_last_seen_v1/init.sql",  # noqa E501
    # HTTP Error 408: Request Time-out
    "sql/moz-fx-data-shared-prod/telemetry_derived/latest_versions/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/italy_covid19_outage_v1/query.sql",
    # Query parameter not found
    "sql/moz-fx-data-shared-prod/telemetry_derived/experiments_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_scalar_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_boolean_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_histogram_aggregates_content_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_histogram_aggregates_parent_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_histogram_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_histogram_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_histogram_bucket_counts_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/glam_client_probe_counts_extract_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/scalar_percentiles_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_scalar_probe_counts_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/asn_aggregates_v1/query.sql",
    # Dataset sql/glam-fenix-dev:glam_etl was not found
    *glob.glob("sql/glam-fenix-dev/glam_etl/**/*.sql", recursive=True),
    # Query templates
    "sql/moz-fx-data-shared-prod/search_derived/mobile_search_clients_daily_v1/fenix_metrics.template.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/search_derived/mobile_search_clients_daily_v1/mobile_search_clients_daily.template.sql",  # noqa
}


class DryRun:
    """Dry run SQL files."""

    # todo: support different dry run endpoints for different projects
    DRY_RUN_URL = (
        "https://us-central1-moz-fx-data-shared-prod.cloudfunctions.net/"
        "bigquery-etl-dryrun"
    )

    def __init__(self, sqlfile):
        """Instantiate DryRun class."""
        self.sqlfile = sqlfile

    @cached_property
    def dry_run_result(self):
        """Dry run the provided SQL file."""
        sql = open(self.sqlfile).read()

        try:
            r = urlopen(
                Request(
                    self.DRY_RUN_URL,
                    headers={"Content-Type": "application/json"},
                    data=json.dumps(
                        {
                            "dataset": basename(dirname(dirname(self.sqlfile))),
                            "query": sql,
                        }
                    ).encode("utf8"),
                    method="POST",
                )
            )
        except Exception as e:
            print(f"{self.sqlfile:59} ERROR\n", e)
            return None

        return json.load(r)

    def get_referenced_tables(self):
        """Return referenced tables by dry running the SQL file."""
        if self.sqlfile not in SKIP and not self.is_valid():
            raise Exception(f"Error when dry running SQL file {self.sqlfile}")

        if self.sqlfile in SKIP:
            print(f"\t...Ignoring dryrun results for {self.sqlfile}")

        if (
            self.dry_run_result
            and self.dry_run_result["valid"]
            and "referencedTables" in self.dry_run_result
        ):
            return self.dry_run_result["referencedTables"]

        return []

    def is_valid(self):
        """Dry run the provided SQL file and check if valid."""
        if self.dry_run_result is None:
            return False

        if "errors" in self.dry_run_result and len(self.dry_run_result["errors"]) == 1:
            error = self.dry_run_result["errors"][0]
        else:
            error = None

        if self.dry_run_result["valid"]:
            print(f"{self.sqlfile:59} OK")
        elif (
            error
            and error.get("code", None) in [400, 403]
            and "does not have bigquery.tables.create permission for dataset"
            in error.get("message", "")
        ):
            # We want the dryrun service to only have read permissions, so
            # we expect CREATE VIEW and CREATE TABLE to throw specific
            # exceptions.
            print(f"{self.sqlfile:59} OK")
        else:
            print(f"{self.sqlfile:59} ERROR\n", self.dry_run_result["errors"])
            return False

        return True


def main():
    """Dry run all SQL files in the project directories."""
    file_names = ("query.sql", "view.sql", "part*.sql")

    sql_files = [
        f
        for n in file_names
        for f in glob.glob(f"sql/**/{n}", recursive=True)
        if f not in SKIP
    ]

    def sql_file_valid(sqlfile):
        """Dry run SQL files."""
        return DryRun(sqlfile).is_valid()

    with ThreadPool(8) as p:
        result = p.map(sql_file_valid, sql_files, chunksize=1)
    if all(result):
        exitcode = 0
    else:
        exitcode = 1
    sys.exit(exitcode)


if __name__ == "__main__":
    main()



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "jklukas@mozilla.com",
    "start_date": datetime.datetime(2019, 3, 1, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "jklukas@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=600),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 1,
}

with DAG(
    "bqetl_fxa_events", default_args=default_args, schedule_interval="30 1 * * *"
) as dag:

    firefox_accounts_derived__exact_mau28__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__exact_mau28__v1",
        destination_table="exact_mau28_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_auth_bounce_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_auth_bounce_events__v1",
        destination_table="fxa_auth_bounce_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        arguments=["--schema_update_option=ALLOW_FIELD_ADDITION"],
        dag=dag,
    )

    firefox_accounts_derived__fxa_auth_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_auth_events__v1",
        destination_table="fxa_auth_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        arguments=["--schema_update_option=ALLOW_FIELD_ADDITION"],
        dag=dag,
    )

    firefox_accounts_derived__fxa_content_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_content_events__v1",
        destination_table="fxa_content_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        arguments=["--schema_update_option=ALLOW_FIELD_ADDITION"],
        dag=dag,
    )

    firefox_accounts_derived__fxa_delete_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_delete_events__v1",
        destination_table="fxa_delete_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_log_auth_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_log_auth_events__v1",
        destination_table="fxa_log_auth_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_log_content_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_log_content_events__v1",
        destination_table="fxa_log_content_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_log_device_command_events__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_log_device_command_events__v1",
        destination_table="fxa_log_device_command_events_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_users_daily__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_users_daily__v1",
        destination_table="fxa_users_daily_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_users_last_seen__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_users_last_seen__v1",
        destination_table="fxa_users_last_seen_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        start_date=datetime.datetime(2019, 4, 23, 0, 0),
        date_partition_parameter="submission_date",
        depends_on_past=True,
        dag=dag,
    )

    firefox_accounts_derived__fxa_users_services_daily__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_users_services_daily__v1",
        destination_table="fxa_users_services_daily_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_users_services_first_seen__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_users_services_first_seen__v1",
        destination_table="fxa_users_services_first_seen_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    firefox_accounts_derived__fxa_users_services_last_seen__v1 = bigquery_etl_query(
        task_id="firefox_accounts_derived__fxa_users_services_last_seen__v1",
        destination_table="fxa_users_services_last_seen_v1",
        dataset_id="firefox_accounts_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        start_date=datetime.datetime(2019, 10, 8, 0, 0),
        date_partition_parameter="submission_date",
        depends_on_past=True,
        dag=dag,
    )

    firefox_accounts_derived__exact_mau28__v1.set_upstream(
        firefox_accounts_derived__fxa_users_last_seen__v1
    )

    firefox_accounts_derived__fxa_users_daily__v1.set_upstream(
        firefox_accounts_derived__fxa_auth_bounce_events__v1
    )

    firefox_accounts_derived__fxa_users_daily__v1.set_upstream(
        firefox_accounts_derived__fxa_auth_events__v1
    )

    firefox_accounts_derived__fxa_users_daily__v1.set_upstream(
        firefox_accounts_derived__fxa_content_events__v1
    )

    firefox_accounts_derived__fxa_users_last_seen__v1.set_upstream(
        firefox_accounts_derived__fxa_users_daily__v1
    )

    firefox_accounts_derived__fxa_users_services_daily__v1.set_upstream(
        firefox_accounts_derived__fxa_auth_events__v1
    )

    firefox_accounts_derived__fxa_users_services_daily__v1.set_upstream(
        firefox_accounts_derived__fxa_content_events__v1
    )

    firefox_accounts_derived__fxa_users_services_first_seen__v1.set_upstream(
        firefox_accounts_derived__fxa_auth_events__v1
    )

    firefox_accounts_derived__fxa_users_services_first_seen__v1.set_upstream(
        firefox_accounts_derived__fxa_content_events__v1
    )

    firefox_accounts_derived__fxa_users_services_last_seen__v1.set_upstream(
        firefox_accounts_derived__fxa_users_services_daily__v1
    )

    firefox_accounts_derived__fxa_users_services_last_seen__v1.set_upstream(
        firefox_accounts_derived__fxa_users_services_first_seen__v1
    )



"""Publish csv files as BigQuery tables."""

import os
import json
from argparse import ArgumentParser
from google.cloud import bigquery

from bigquery_etl.util.common import project_dirs

DATA_FILENAME = "data.csv"
SCHEMA_FILENAME = "schema.json"
DESCRIPTION_FILENAME = "description.txt"


def _parse_args():
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "--project-id", "--project_id", help="Project to publish tables to"
    )
    return parser.parse_args()


def _load_table(
    data_file_path, schema_file_path=None, description_file_path=None, project=None
):
    client = bigquery.Client(project)

    # Assume path is ...project/dataset/table/data.csv
    path_split = os.path.normcase(data_file_path).split("/")
    dataset_id = path_split[-3]
    table_id = path_split[-2]
    if not project:
        project = path_split[0]
    dataset_ref = client.dataset(dataset_id, project=project)
    table_ref = dataset_ref.table(table_id)

    job_config = bigquery.LoadJobConfig(
        source_format=bigquery.SourceFormat.CSV,
        skip_leading_rows=1,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
    )
    with open(data_file_path, "rb") as data_file:
        if schema_file_path is None:
            fields = data_file.readline().decode().strip().split(",")
            # Assume all fields are strings and nullable
            job_config.schema = [
                bigquery.SchemaField(field, field_type="STRING") for field in fields
            ]
            data_file.seek(0)
        else:
            with open(schema_file_path) as schema_file:
                fields = json.load(schema_file)
            job_config.schema = [
                bigquery.SchemaField(
                    field["name"],
                    field_type=field.get("type", "STRING"),
                    mode=field.get("mode", "NULLABLE"),
                    description=field.get("description"),
                )
                for field in fields
            ]

        job = client.load_table_from_file(data_file, table_ref, job_config=job_config)

    job.result()

    if description_file_path is not None:
        with open(description_file_path) as description_file:
            description = description_file.read()
            table = client.get_table(table_ref)
            table.description = description
            client.update_table(table, ["description"])


def main():
    """Publish csv files as BigQuery tables."""
    args = _parse_args()
    projects = project_dirs(args.project_id)

    for data_dir in projects:
        for root, dirs, files in os.walk(data_dir):
            for filename in files:
                if filename == DATA_FILENAME:
                    schema_file_path = (
                        os.path.join(root, SCHEMA_FILENAME)
                        if SCHEMA_FILENAME in files
                        else None
                    )
                    description_file_path = (
                        os.path.join(root, DESCRIPTION_FILENAME)
                        if DESCRIPTION_FILENAME in files
                        else None
                    )
                    _load_table(
                        os.path.join(root, filename),
                        schema_file_path,
                        description_file_path,
                        args.project_id,
                    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Experiment enrollment aggregates live view query generator."""
import argparse
import datetime
import json
import sys
import textwrap
from time import sleep


p = argparse.ArgumentParser()
p.add_argument(
    "--submission-date",
    type=str,
    help="Cut-off date for using pre-computed vs live tables in view",
    required=True
)
p.add_argument(
    "--json-output",
    action='store_true',
    help="Output the result wrapped in json parseable as an XCOM",
)
p.add_argument(
    "--wait-seconds",
    type=int,
    default=0,
    help="Add a delay before executing the script to allow time for the xcom sidecar to complete startup",
)

def generate_sql(opts):
    """Create a BigQuery SQL query for the experiment enrollments aggregates view with new date filled in.
       Unfortunately, BigQuery does not allow parameters in view definitions, so this script is a very thin
       wrapper over the nearly complete sql query to fill in the cutoff date for using the pre-aggregated vs
       live table for the view"""
    query = textwrap.dedent(
        """
        CREATE
        OR REPLACE VIEW `moz-fx-data-shared-prod.telemetry_derived.experiment_enrollment_aggregates_live` AS
          WITH live AS (
            SELECT
              event_object AS `type`,
              event_string_value AS experiment,
              `moz-fx-data-shared-prod`.udf.get_key(event_map_values, 'branch') AS branch,
              TIMESTAMP_ADD(
                TIMESTAMP_TRUNC(`timestamp`, HOUR),
                -- Aggregates event counts over 5-minute intervals
                INTERVAL (DIV(EXTRACT(MINUTE FROM `timestamp`), 5) * 5) MINUTE
              ) AS window_start,
              TIMESTAMP_ADD(
                TIMESTAMP_TRUNC(`timestamp`, HOUR),
                INTERVAL ((DIV(EXTRACT(MINUTE FROM `timestamp`), 5) + 1) * 5) MINUTE
              ) AS window_end,
              COUNTIF(event_method = 'enroll') AS enroll_count,
              COUNTIF(event_method = 'unenroll') AS unenroll_count,
              COUNTIF(event_method = 'graduate') AS graduate_count,
              COUNTIF(event_method = 'update') AS update_count,
              COUNTIF(event_method = 'enrollFailed') AS enroll_failed_count,
              COUNTIF(event_method = 'unenrollFailed') AS unenroll_failed_count,
              COUNTIF(event_method = 'updateFailed') AS update_failed_count
            FROM
              `moz-fx-data-shared-prod.telemetry_derived.events_live`
            WHERE
              event_category = 'normandy'
              AND date(timestamp) > '{submission_date}'
          GROUP BY
            `type`,
            experiment,
            branch,
            window_start,
            window_end
        ),
        previous AS (
          SELECT
            *
          FROM
            `moz-fx-data-shared-prod.telemetry_derived.experiment_enrollment_aggregates_v1`
          WHERE
            date(window_start) <= '{submission_date}'
        ),
        all_enrollments AS (
          SELECT
            *
          FROM
            previous
          UNION ALL
          SELECT
            *
          FROM
            live
        )
        SELECT
          *,
          SUM(enroll_count) OVER previous_rows_window AS cumulative_enroll_count,
          SUM(unenroll_count) OVER previous_rows_window AS cumulative_unenroll_count,
          SUM(graduate_count) OVER previous_rows_window AS cumulative_graduate_count,
          SUM(update_count) OVER previous_rows_window AS cumulative_update_count,
          SUM(enroll_failed_count) OVER previous_rows_window AS cumulative_enroll_failed_count,
          SUM(unenroll_failed_count) OVER previous_rows_window AS cumulative_unenroll_failed_count,
          SUM(update_failed_count) OVER previous_rows_window AS cumulative_update_failed_count
        FROM
          all_enrollments
        WINDOW previous_rows_window AS (
          PARTITION BY
            experiment,
            branch
          ORDER BY
            window_start
          ROWS BETWEEN
            UNBOUNDED PRECEDING
            AND CURRENT ROW
        )
        """.format(
            **opts
        )
    )
    if opts['json_output']:
        return json.dumps(query)
    else:
        return query


def main(argv, out=print):
    """Print experiment enrollments aggregates view sql to stdout."""
    opts = vars(p.parse_args(argv[1:]))
    sleep(opts['wait_seconds'])
    out(generate_sql(opts))


if __name__ == "__main__":
    main(sys.argv)



"""
Copy a day's data from live to stable ping tables, deduplicating on document_id.

By default, the script will process all tables in datasets named
like *_live, copying data into table of the same name in *_stable
datasets. The script can be configured to exclude a list of tables
or to process only a specific list of tables.
"""

from argparse import ArgumentParser
from datetime import datetime, timedelta
from itertools import groupby
from multiprocessing.pool import ThreadPool
from uuid import uuid4
import logging
from google.cloud import bigquery

from bigquery_etl.util import standard_args
from bigquery_etl.util.client_queue import ClientQueue
from bigquery_etl.util.bigquery_id import sql_table_id


QUERY_TEMPLATE = """
WITH
  -- Distinct document_ids and their minimum submission_timestamp today
  -- not including document_ids that only occur on or after @end_time
  distinct_document_ids AS (
  SELECT
    document_id,
    MIN(submission_timestamp) AS submission_timestamp
  FROM
    `{live_table}`
  WHERE
    DATE(submission_timestamp) >= DATE_SUB(
        DATE(@start_time),
        INTERVAL @num_preceding_days DAY
    )
    AND submission_timestamp < @end_time
    -- Bug 1657360
    AND 'automation' NOT IN (
      SELECT TRIM(t) FROM UNNEST(SPLIT(metadata.header.x_source_tags, ',')) t
    )
  GROUP BY
    document_id
  HAVING
    submission_timestamp >= @start_time),
  -- A single slice of a live ping table.
  base AS (
  SELECT
    *
  FROM
    `{live_table}`
  JOIN
    distinct_document_ids
  USING
    -- Retain only the first seen documents for each ID, according to timestamp.
    (document_id, submission_timestamp)
  WHERE
    submission_timestamp >= @start_time
    AND submission_timestamp < @end_time),
  --
  -- Order documents by assigning a row number.
  numbered_duplicates AS (
  SELECT
    *,
    ROW_NUMBER() OVER (PARTITION BY document_id) AS _n
  FROM
    base)
  --
  -- Retain only one document for each ID.
SELECT
  * EXCEPT(_n)
FROM
  numbered_duplicates
WHERE
  _n = 1
"""

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--project_id",
    "--project-id",
    default="moz-fx-data-shar-nonprod-efed",
    help="ID of the project in which to find tables",
)
parser.add_argument(
    "--dates",
    "--date",
    nargs="+",
    required=True,
    type=lambda d: datetime.strptime(d, "%Y-%m-%d").date(),
    help="One or more days of data to copy, in format 2019-01-01",
)
standard_args.add_parallelism(parser)
standard_args.add_dry_run(parser, debug_log_queries=False)
standard_args.add_log_level(parser)
standard_args.add_priority(parser)
parser.add_argument(
    "--slices",
    type=int,
    default=1,
    help=(
        "Number of queries to split deduplicate over, each handling an equal-size time "
        "slice of the date; avoids memory overflow at the cost of less effective "
        "clustering; recommended only for tables failing due to memory overflow"
    ),
)
parser.add_argument(
    "--hourly",
    action="store_const",
    dest="slices",
    const=24,
    help="Deduplicate one hour at a time; equivalent to --slices=24",
)
parser.add_argument(
    "--preceding_days",
    "--preceding-days",
    type=int,
    default=0,
    help="Number of days preceding --date that should be used to filter out duplicates",
)
standard_args.add_billing_projects(parser)
standard_args.add_table_filter(parser)

temporary_dataset = None


def get_temporary_dataset(client):
    """Get a cached reference to the dataset used for server-assigned destinations."""
    global temporary_dataset
    if temporary_dataset is None:
        # look up the dataset used for query results without a destination
        dry_run = bigquery.QueryJobConfig(dry_run=True)
        destination = client.query("SELECT 1", dry_run).destination
        temporary_dataset = client.dataset(destination.dataset_id, destination.project)
    return temporary_dataset


def get_temporary_table(client):
    """Generate a temporary table and return the specified date partition.

    Generates a table name that looks similar to, but won't collide with, a
    server-assigned table and that the web console will consider temporary.
    Table expiration can't be set from a query job, so the table is created
    here.

    Query destination must be explicity defined in order for results to use
    partitioning, and by extension clustering. Additionally, the only way to
    create a temporary table with both expiration and time partitioning as a
    result of a query is to use a CREATE TABLE statement.

    Query destination must be generated locally and never collide with
    server-assigned table names, because server-assigned tables cannot be
    modified. Server-assigned tables for a dry_run query cannot be detected by
    client.list_tables and cannot be reused as that constitutes a modification.

    Server-assigned tables have names that start with "anon" and follow with
    either 40 hex characters or a uuid replacing "-" with "_", and cannot be
    modified (i.e. reused).

    The web console considers a table temporary if the dataset name starts with
    "_" and table_id starts with "anon" and is followed by at least one
    character.
    """
    dataset = get_temporary_dataset(client)
    return sql_table_id(dataset.table(f"anon{uuid4().hex}"))


def _get_query_job_configs(
    client, live_table, date, dry_run, slices, priority, preceding_days
):
    sql = QUERY_TEMPLATE.format(live_table=live_table)
    stable_table = f"{live_table.replace('_live.', '_stable.', 1)}${date:%Y%m%d}"
    kwargs = dict(use_legacy_sql=False, dry_run=dry_run, priority=priority)
    start_time = datetime(*date.timetuple()[:6])
    end_time = start_time + timedelta(days=1)
    if slices > 1:
        # create temporary tables with stable_table's time_partitioning and
        # clustering_fields, and a 1-day expiration
        stable_table = client.get_table(stable_table)
        ddl = "CREATE TABLE\n  `{dest}`"
        ddl += f"\nPARTITION BY\n  DATE({stable_table.time_partitioning.field})"
        if stable_table.clustering_fields:
            ddl += f"\nCLUSTER BY\n  {', '.join(stable_table.clustering_fields)}"
        ddl += "\nOPTIONS\n  (expiration_timestamp = "
        ddl += "TIMESTAMP_ADD(CURRENT_TIMESTAMP, INTERVAL 1 DAY))"
        slice_size = (end_time - start_time) / slices
        params = [start_time + slice_size * i for i in range(slices)] + [
            end_time
        ]  # explicitly use end_time to avoid rounding errors
        return [
            (
                f"{ddl.format(dest=get_temporary_table(client))}\nAS\n{sql.strip()}",
                stable_table,
                bigquery.QueryJobConfig(
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            "start_time", "TIMESTAMP", params[i]
                        ),
                        bigquery.ScalarQueryParameter(
                            "end_time", "TIMESTAMP", params[i + 1]
                        ),
                        bigquery.ScalarQueryParameter(
                            "num_preceding_days", "INT64", preceding_days
                        ),
                    ],
                    **kwargs,
                ),
            )
            for i in range(slices)
        ]
    else:
        return [
            (
                sql,
                stable_table,
                bigquery.QueryJobConfig(
                    destination=stable_table,
                    write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            "start_time", "TIMESTAMP", start_time
                        ),
                        bigquery.ScalarQueryParameter(
                            "end_time", "TIMESTAMP", end_time
                        ),
                        bigquery.ScalarQueryParameter(
                            "num_preceding_days", "INT64", preceding_days
                        ),
                    ],
                    **kwargs,
                ),
            )
        ]


def _run_deduplication_query(client, sql, stable_table, job_config):
    query_job = client.query(sql, job_config)
    if not query_job.dry_run:
        query_job.result()
    return stable_table, query_job


def _copy_join_parts(client, stable_table, query_jobs):
    total_bytes = sum(query.total_bytes_processed for query in query_jobs)
    if query_jobs[0].dry_run:
        api_repr = query_jobs[0].to_api_repr()
        if len(query_jobs) > 1:
            logging.info(f"Would process {total_bytes} bytes: [{api_repr},...]")
            logging.info(f"Would copy {len(query_jobs)} results to {stable_table}")
        else:
            logging.info(f"Would process {total_bytes} bytes: {api_repr}")
    else:
        logging.info(f"Processed {total_bytes} bytes to populate {stable_table}")
        if len(query_jobs) > 1:
            partition_id = stable_table.table_id.split("$", 1)[1]
            sources = [
                f"{sql_table_id(job.destination)}${partition_id}" for job in query_jobs
            ]
            job_config = bigquery.CopyJobConfig(
                write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE
            )
            copy_job = client.copy_table(sources, stable_table, job_config=job_config)
            copy_job.result()
            logging.info(f"Copied {len(query_jobs)} results to populate {stable_table}")
            for job in query_jobs:
                client.delete_table(job.destination)
            logging.info(f"Deleted {len(query_jobs)} temporary tables")


def _contains_glob(patterns):
    return any(set("*?[").intersection(pattern) for pattern in patterns)


def _glob_dataset(pattern):
    return pattern.split(".", 1)[0]


def _list_live_tables(client, pool, project_id, only_tables, table_filter):
    if only_tables and not _contains_glob(only_tables):
        # skip list calls when only_tables exists and contains no globs
        return [f"{project_id}.{t}" for t in only_tables if table_filter(t)]
    if only_tables and not _contains_glob(_glob_dataset(t) for t in only_tables):
        # skip list_datasets call when only_tables exists and datasets contain no globs
        live_datasets = {f"{project_id}.{_glob_dataset(t)}" for t in only_tables}
    else:
        live_datasets = [
            d.reference
            for d in client.list_datasets(project_id)
            if d.dataset_id.endswith("_live")
        ]
    return [
        sql_table_id(t)
        for tables in pool.map(client.list_tables, live_datasets)
        for t in tables
        if table_filter(f"{t.dataset_id}.{t.table_id}")
        and "beam_load_sink" not in t.table_id
    ]


def main():
    """Copy a day's data from live to stable ping tables, dedup on document_id."""
    args = parser.parse_args()

    # create a queue for balancing load across projects
    client_q = ClientQueue(args.billing_projects, args.parallelism)

    with ThreadPool(args.parallelism) as pool:
        with client_q.client() as client:
            live_tables = _list_live_tables(
                client=client,
                pool=pool,
                project_id=args.project_id,
                only_tables=getattr(args, "only_tables", None),
                table_filter=args.table_filter,
            )

            query_jobs = [
                (_run_deduplication_query, *args)
                for jobs in pool.starmap(
                    _get_query_job_configs,
                    [
                        (
                            client,  # only use one client to create temp tables
                            live_table,
                            date,
                            args.dry_run,
                            args.slices,
                            args.priority,
                            args.preceding_days,
                        )
                        for live_table in live_tables
                        for date in args.dates
                    ],
                )
                for args in jobs
            ]

        # preserve query_jobs order so results stay sorted by stable_table for groupby
        results = pool.starmap(client_q.with_client, query_jobs, chunksize=1)
        copy_jobs = [
            (_copy_join_parts, stable_table, [query_job for _, query_job in group])
            for stable_table, group in groupby(results, key=lambda result: result[0])
        ]
        pool.starmap(client_q.with_client, copy_jobs, chunksize=1)


if __name__ == "__main__":
    main()



"""
Generate one view definition file per document type in '_stable' tables.

If there are existing view definitions in the destination directory then those will be
kept instead.

Run as:
  ./script/generate_views 'moz-fx-data-shared-prod:*_stable.*'
"""

import logging
import os
import re
from argparse import ArgumentParser
from fnmatch import fnmatchcase

from google.cloud import bigquery

from bigquery_etl.util import standard_args
from bigquery_etl.util.bigquery_tables import get_tables_matching_patterns

VERSION_RE = re.compile(r"_v([0-9]+)$")
WHITESPACE_RE = re.compile(r"\s+")
WILDCARD_RE = re.compile(r"[*?[]")

DEFAULT_PATTERN = "telemetry.*"
DEFAULT_EXCLUDE = r"*_raw"

VIEW_QUERY_TEMPLATE = """
SELECT
  * REPLACE(
    {replacements})
FROM
  `{target}`
"""

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "patterns",
    metavar="[project:]dataset[.table]",
    default=[DEFAULT_PATTERN],
    nargs="*",
    help="Table that should have a latest-version view, may use shell-style wildcards,"
    f" defaults to: {DEFAULT_PATTERN}",
)
parser.add_argument(
    "--exclude",
    action="append",
    default=[DEFAULT_EXCLUDE],
    metavar="project:dataset.table",
    help="Latest-version views that should be ignored, may use shell-style wildcards,"
    f" defaults to: {DEFAULT_EXCLUDE}",
)
parser.add_argument(
    "--sql-dir", default="sql/", help="The path where generated SQL files are stored."
)
standard_args.add_log_level(parser)


def main():
    """Generate view definitions."""
    args = parser.parse_args()

    # set log level
    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        parser.error(f"argument --log-level: {e}")

    client = bigquery.Client()
    tables = get_tables_matching_patterns(client, args.patterns)

    views = {}
    for full_table_id in tables:
        view = VERSION_RE.sub("", full_table_id)
        if view not in views:
            views[view] = []
        views[view].append(full_table_id)

    create_views_if_not_exist(client, views, args.exclude, args.sql_dir)


def create_views_if_not_exist(client, views, exclude, sql_dir):
    """Create views unless a local file for creating the view exists."""
    for view, tables in views.items():
        if any(fnmatchcase(pattern, view) for pattern in exclude):
            logging.info("skipping table: matched by exclude pattern: {view}")
            continue
        if view.endswith("_"):
            # A trailing '_' confuses the logic here of parsing versions,
            # and likely indicates that the table is somehow private, so
            # we ignore it.
            logging.info("skipping table ending in _: {view}")
            continue

        version = max(
            int(match.group()[2:])
            for table in tables
            for match in [VERSION_RE.search(table)]
            if match is not None
        )

        project, dataset, viewname = view.split(".")
        target = f"{view}_v{version}"
        view_dataset = dataset.rsplit("_", 1)[0]
        full_view_id = ".".join([project, view_dataset, viewname])
        target_file = os.path.join(sql_dir, project, view_dataset, viewname, "view.sql")

        if not os.path.exists(target_file):
            # We put this BQ API all inside the conditional to speed up execution
            # in the case target files already exist.
            table = client.get_table(target)
            replacements = ["mozfun.norm.metadata(metadata)" " AS metadata"]
            # Add special replacements for Glean pings.
            schema_id = table.labels.get("schema_id", None)
            if schema_id == "glean_ping_1":
                replacements += [
                    "mozfun.norm.glean_ping_info(ping_info)" " AS ping_info"
                ]
                if table.table_id == "baseline_v1":
                    replacements += [
                        "mozfun.norm.glean_baseline_client_info"
                        "(client_info, metrics)"
                        " AS client_info"
                    ]
                if table.dataset_id.startswith(
                    "org_mozilla_fenix"
                ) and table.table_id.startswith("metrics"):
                    # todo: use mozfun udfs
                    replacements += [
                        "`moz-fx-data-shared-prod.udf.normalize_fenix_metrics`"
                        "(client_info.telemetry_sdk_build, metrics)"
                        " AS metrics"
                    ]
                if table.dataset_id.startswith("firefox_desktop"):
                    # Bug 1672191
                    replacements += [
                        "'Firefox' AS normalized_app_name",
                    ]
            elif schema_id in ("main_ping_1", "main_ping_4"):
                replacements += [
                    "`moz-fx-data-shared-prod.udf.normalize_main_payload`(payload)"
                    " AS payload"
                ]
            replacements = ",\n    ".join(replacements)
            view_query = VIEW_QUERY_TEMPLATE.format(
                target=target, replacements=replacements
            ).strip()
            full_sql = f"CREATE OR REPLACE VIEW\n  `{full_view_id}`\nAS {view_query}\n"
            print("Creating " + target_file)
            if not os.path.exists(os.path.dirname(target_file)):
                os.makedirs(os.path.dirname(target_file))
            with open(target_file, "w") as f:
                f.write(full_sql)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Compare number of document IDs in structured decoded, live, and stable tables."""

import datetime
from argparse import ArgumentParser
from collections import defaultdict

from google.cloud import bigquery

DECODED_QUERY = """
WITH decoded AS (
  SELECT 
    * EXCEPT (metadata),  -- Some tables have different field order in metadata
    _TABLE_SUFFIX,
  FROM
    `moz-fx-data-shared-prod.payload_bytes_decoded.structured_*`
  UNION ALL
  SELECT
    * EXCEPT (metadata),
    _TABLE_SUFFIX,
  FROM
    `moz-fx-data-shared-prod.payload_bytes_decoded.stub_installer_*` 
)
SELECT
  DATE(submission_timestamp) AS submission_date,
  SPLIT(_TABLE_SUFFIX, '__')[OFFSET(0)] AS namespace,
  SPLIT(_TABLE_SUFFIX, '__')[OFFSET(1)] AS doc_type,
  COUNT(DISTINCT(document_id)) AS docid_count,
FROM
  decoded
WHERE
  DATE(submission_timestamp) = '{date}'
GROUP BY
  namespace,
  doc_type,
  submission_date
"""

LIVE_STABLE_QUERY = """
SELECT
  DATE(submission_timestamp) AS submission_date,
  _TABLE_SUFFIX AS doc_type,
  COUNT(DISTINCT(document_id)) AS docid_count,
FROM
  `moz-fx-data-shared-prod.{namespace}_{type}.*`
WHERE
  DATE(submission_timestamp) = '{date}'
GROUP BY
  doc_type,
  submission_date
"""

EXCLUDED_NAMESPACES = {"xfocsp_error_report"}  # restricted access


def get_docid_counts(date, project, destination_dataset, destination_table):
    """Get distinct docid count for decoded, live, and stable tables.

    Results are saved to bigquery where each row contains the decoded, live,
    and stable docid counts for each namespace and doc type combination.
    """
    client = bigquery.Client(project=project)

    # key is tuple of (namespace, doc_type), value is dict where key is
    # table type (live, stable, decoded) and value is docid count
    docid_counts_by_doc_type_by_table = defaultdict(dict)

    namespaces = set()

    print("Getting decoded table doc id counts")
    decoded_query_results = client.query(DECODED_QUERY.format(date=date)).result()

    # Get docid counts in decoded tables
    for row in decoded_query_results:
        if row["namespace"] in EXCLUDED_NAMESPACES:
            continue

        docid_counts_by_doc_type_by_table[(row["namespace"], row["doc_type"])][
            "decoded"
        ] = row["docid_count"]

        namespaces.add(row["namespace"])

    # Get docid counts in stable and live tables
    for namespace in namespaces:
        print(f"Getting {namespace} stable doc id counts")
        stable_query_results = client.query(
            LIVE_STABLE_QUERY.format(date=date, namespace=namespace, type="stable")
        ).result()

        for row in stable_query_results:
            docid_counts_by_doc_type_by_table[(namespace, row["doc_type"])][
                "stable"
            ] = row["docid_count"]

        print(f"Getting {namespace} live doc id counts")
        live_query_results = client.query(
            LIVE_STABLE_QUERY.format(date=date, namespace=namespace, type="live")
        ).result()

        for row in live_query_results:
            docid_counts_by_doc_type_by_table[(namespace, row["doc_type"])][
                "live"
            ] = row["docid_count"]

    # Transform into format for bigquery load
    output_data = []
    for (namespace, doctype), data in docid_counts_by_doc_type_by_table.items():
        data["submission_date"] = str(date)
        data["namespace"] = namespace
        data["doc_type"] = doctype
        output_data.append(data)

    load_config = bigquery.LoadJobConfig(
        time_partitioning=bigquery.TimePartitioning(field="submission_date"),
        create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
    )

    load_job = client.load_table_from_json(
        json_rows=output_data,
        destination=f"{destination_dataset}.{destination_table}"
        f"${date.strftime('%Y%m%d')}",
        job_config=load_config,
    )

    load_job.result()


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", type=datetime.date.fromisoformat, required=True)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--destination-dataset", default="monitoring")
    parser.add_argument("--destination-table", default="structured_distinct_docids_v1")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    get_docid_counts(
        args.date, args.project, args.destination_dataset, args.destination_table,
    )



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "dthorn@mozilla.com",
    "start_date": datetime.datetime(2020, 10, 8, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "dthorn@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_mozilla_vpn", default_args=default_args, schedule_interval="@daily"
) as dag:

    mozilla_vpn_derived__add_device_events__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_derived__add_device_events__v1",
        destination_table="add_device_events_v1",
        dataset_id="mozilla_vpn_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    mozilla_vpn_derived__login_flows__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_derived__login_flows__v1",
        destination_table="login_flows_v1",
        dataset_id="mozilla_vpn_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    mozilla_vpn_derived__protected__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_derived__protected__v1",
        destination_table="protected_v1",
        dataset_id="mozilla_vpn_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    mozilla_vpn_derived__users__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_derived__users__v1",
        destination_table="users_v1",
        dataset_id="mozilla_vpn_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    mozilla_vpn_derived__waitlist__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_derived__waitlist__v1",
        destination_table="waitlist_v1",
        dataset_id="mozilla_vpn_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    mozilla_vpn_external__devices__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_external__devices__v1",
        destination_table="devices_v1",
        dataset_id="mozilla_vpn_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=True,
        parameters=[
            "external_database_query:STRING:SELECT * FROM devices WHERE DATE(updated_at) = DATE '{{ds}}'"
        ],
        dag=dag,
    )

    mozilla_vpn_external__subscriptions__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_external__subscriptions__v1",
        destination_table="subscriptions_v1",
        dataset_id="mozilla_vpn_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=True,
        parameters=[
            "external_database_query:STRING:SELECT * FROM subscriptions WHERE DATE(updated_at) = DATE '{{ds}}'"
        ],
        dag=dag,
    )

    mozilla_vpn_external__users__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_external__users__v1",
        destination_table="users_v1",
        dataset_id="mozilla_vpn_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=True,
        parameters=[
            "external_database_query:STRING:SELECT * FROM users WHERE DATE(updated_at) = DATE '{{ds}}'"
        ],
        dag=dag,
    )

    mozilla_vpn_external__waitlist__v1 = bigquery_etl_query(
        task_id="mozilla_vpn_external__waitlist__v1",
        destination_table="waitlist_v1",
        dataset_id="mozilla_vpn_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=True,
        parameters=[
            "external_database_query:STRING:SELECT * FROM vpn_waitlist WHERE DATE(updated_at) = DATE '{{ds}}'"
        ],
        dag=dag,
    )

    wait_for_firefox_accounts_derived__fxa_auth_events__v1 = ExternalTaskSensor(
        task_id="wait_for_firefox_accounts_derived__fxa_auth_events__v1",
        external_dag_id="bqetl_fxa_events",
        external_task_id="firefox_accounts_derived__fxa_auth_events__v1",
        execution_delta=datetime.timedelta(days=-1, seconds=81000),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    mozilla_vpn_derived__login_flows__v1.set_upstream(
        wait_for_firefox_accounts_derived__fxa_auth_events__v1
    )
    wait_for_firefox_accounts_derived__fxa_content_events__v1 = ExternalTaskSensor(
        task_id="wait_for_firefox_accounts_derived__fxa_content_events__v1",
        external_dag_id="bqetl_fxa_events",
        external_task_id="firefox_accounts_derived__fxa_content_events__v1",
        execution_delta=datetime.timedelta(days=-1, seconds=81000),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    mozilla_vpn_derived__login_flows__v1.set_upstream(
        wait_for_firefox_accounts_derived__fxa_content_events__v1
    )

    mozilla_vpn_derived__users__v1.set_upstream(mozilla_vpn_external__users__v1)

    mozilla_vpn_derived__waitlist__v1.set_upstream(mozilla_vpn_external__waitlist__v1)



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from operators.gcp_container_operator import GKEPodOperator
from utils.gcp import gke_command

default_args = {
    "owner": "ascholtz@mozilla.com",
    "start_date": datetime.datetime(2020, 4, 14, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "ascholtz@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_public_data_json", default_args=default_args, schedule_interval="0 4 * * *"
) as dag:
    docker_image = "mozilla/bigquery-etl:latest"

    export_public_data_json_telemetry_derived__deviations__v1 = GKEPodOperator(
        task_id="export_public_data_json_telemetry_derived__deviations__v1",
        name="export_public_data_json_telemetry_derived__deviations__v1",
        arguments=["script/publish_public_data_json"]
        + [
            "--query_file=sql/moz-fx-data-shared-prod/telemetry_derived/deviations_v1/query.sql"
        ]
        + ["--destination_table=deviations${{ds_nodash}}"]
        + ["--dataset_id=telemetry_derived"]
        + ["--project_id=moz-fx-data-shared-prod"]
        + ["--parameter=submission_date:DATE:{{ds}}"],
        image=docker_image,
        dag=dag,
    )

    export_public_data_json_telemetry_derived__ssl_ratios__v1 = GKEPodOperator(
        task_id="export_public_data_json_telemetry_derived__ssl_ratios__v1",
        name="export_public_data_json_telemetry_derived__ssl_ratios__v1",
        arguments=["script/publish_public_data_json"]
        + [
            "--query_file=sql/moz-fx-data-shared-prod/telemetry_derived/ssl_ratios_v1/query.sql"
        ]
        + ["--destination_table=ssl_ratios${{ds_nodash}}"]
        + ["--dataset_id=telemetry_derived"]
        + ["--project_id=moz-fx-data-shared-prod"]
        + ["--parameter=submission_date:DATE:{{ds}}"],
        image=docker_image,
        dag=dag,
    )

    wait_for_telemetry_derived__deviations__v1 = ExternalTaskSensor(
        task_id="wait_for_telemetry_derived__deviations__v1",
        external_dag_id="bqetl_deviations",
        external_task_id="telemetry_derived__deviations__v1",
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    export_public_data_json_telemetry_derived__deviations__v1.set_upstream(
        wait_for_telemetry_derived__deviations__v1
    )

    wait_for_telemetry_derived__ssl_ratios__v1 = ExternalTaskSensor(
        task_id="wait_for_telemetry_derived__ssl_ratios__v1",
        external_dag_id="bqetl_ssl_ratios",
        external_task_id="telemetry_derived__ssl_ratios__v1",
        execution_delta=datetime.timedelta(seconds=7200),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    export_public_data_json_telemetry_derived__ssl_ratios__v1.set_upstream(
        wait_for_telemetry_derived__ssl_ratios__v1
    )

    public_data_gcs_metadata = gke_command(
        task_id="public_data_gcs_metadata",
        command=["script/publish_public_data_gcs_metadata"],
        docker_image=docker_image,
        dag=dag,
    )

    public_data_gcs_metadata.set_upstream(
        [
            export_public_data_json_telemetry_derived__deviations__v1,
            export_public_data_json_telemetry_derived__ssl_ratios__v1,
        ]
    )



"""Import Stripe data into BigQuery."""

from datetime import datetime, timedelta, timezone
from hashlib import sha256
from tempfile import TemporaryFile
from typing import Any, Dict, IO, List, Optional, Type
import os.path
import re
import sys
import ujson
import warnings

from google.cloud import bigquery
from stripe.api_resources.abstract import ListableAPIResource
import click
import stripe

# event data types with separate events and a defined schema
EVENT_DATA_TYPES = (
    stripe.Charge,
    stripe.CreditNote,
    stripe.Customer,
    stripe.Dispute,
    stripe.Invoice,
    stripe.PaymentIntent,
    stripe.Payout,
    stripe.Plan,
    stripe.Price,
    stripe.Product,
    stripe.SetupIntent,
    stripe.Subscription,
)


class StripeResourceType(click.ParamType):
    """Click parameter type for stripe listable resources."""

    name = "stripe resource"

    def convert(self, value, param, ctx):
        """Get a listable stripe resource type by name."""
        if not isinstance(value, str):
            self.fail(f"{value!r} is not a string")
        if value.islower():
            value = value.capitalize()
        try:
            result = getattr(stripe, value)
        except AttributeError:
            self.fail(f"resource type {value!r} not found in stripe")
        if not issubclass(result, ListableAPIResource):
            self.fail("resource must be listable")
        return result


def bigquery_format(obj: Any, *path):
    """Format stripe objects for BigQuery."""
    if isinstance(obj, stripe.ListObject):
        if obj.data and isinstance(obj.data[0], EVENT_DATA_TYPES):
            # don't expand lists of types that get updated in separate events
            return None
        # recursively request any additional values for paged lists.
        return [
            bigquery_format(e, *path, i) for i, e in enumerate(obj.auto_paging_iter())
        ]
    if path[-1:] == ("metadata",) and isinstance(obj, dict):
        if "userid" in obj:
            # hash fxa uid before it reaches BigQuery
            obj["fxa_uid"] = sha256(obj.pop("userid").encode()).hexdigest()
        # format metadata as a key-value list
        return [
            {"key": key, "value": bigquery_format(value, *path, key)}
            for key, value in obj.items()
        ]
    if isinstance(obj, dict):
        # recursively format and drop nulls, empty lists, and empty objects
        return {
            key: formatted
            for key, value in obj.items()
            # drop use_stripe_sdk because the contents are only for use in Stripe.js
            # https://stripe.com/docs/api/payment_intents/object#payment_intent_object-next_action-use_stripe_sdk
            if key != "use_stripe_sdk"
            for formatted in (bigquery_format(value, *path, key),)
            # drop nulls, empty lists, and empty objects
            if formatted not in (None, [], {})
        }
    return obj


def _open_file(
    api_key: Optional[str], file: Optional[str], table: Optional[str]
) -> IO[bytes]:
    if file is not None:
        if api_key is None:
            mode = "rb"
        else:
            mode = "w+b"
        return open(file, mode=mode)
    elif table is None:
        return sys.stdout.buffer
    elif api_key is None:
        return sys.stdin.buffer
    else:
        return TemporaryFile(mode="w+b")


def _get_schema(
    resource: Type[ListableAPIResource],
) -> Optional[List[bigquery.SchemaField]]:
    snake_case = re.sub(r"(?<!^)(?=[A-Z])", "_", resource.__name__).lower()
    path = os.path.join(os.path.dirname(__file__), f"{snake_case}.schema.json")
    if not os.path.exists(path):
        return None
    with open(path) as fp:
        return bigquery.SchemaField.from_api_repr(
            {"name": "root", "type": "RECORD", "fields": ujson.load(fp)}
        ).fields


@click.group(help="Commands for Stripe ETL.")
def stripe_():
    """Create the CLI group for stripe commands."""
    pass


@stripe_.command("import", help=__doc__)
@click.option(
    "--api-key",
    help="Stripe API key to use for authentication; If not set resources will be read "
    "from stdin, or --file if set",
)
@click.option(
    "--date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Creation date of resources to pull from stripe API; Added to table if set "
    "to ensure only that date partition is replaced",
)
@click.option(
    "--table",
    help="BigQuery Standard SQL format table ID where resources will be written; "
    "if not set resources will be writtent to stdout, or --file if set",
)
@click.option(
    "--file",
    help="Optional persistent file where resources are or will be stored in newline "
    "delimited json format for uploading to BigQuery",
)
@click.option(
    "--resource",
    default="Event",
    type=StripeResourceType(),
    help="Type of stripe resource to export",
)
def import_(
    api_key: str,
    date: datetime,
    table: str,
    file: str,
    resource: Type[ListableAPIResource],
):
    """Import Stripe data into BigQuery."""
    if resource is stripe.Event and not date:
        click.echo("must specify --date for --resource=Event")
        sys.exit(1)
    if api_key is None and table is None:
        click.echo("must specify --api-key and/or --table")
        sys.exit(1)
    if table and date:
        table = f"{table}${date:%Y%m%d}"
    with _open_file(api_key, file, table) as file_obj:
        if api_key:
            stripe.api_key = api_key
            kwargs: Dict[str, Any] = {}
            if date:
                start = date.replace(tzinfo=timezone.utc)
                kwargs["created"] = {
                    "gte": int(start.timestamp()),
                    # make sure to use timedelta before converting to timestamp,
                    # so that leap seconds are properly accounted for.
                    "lt": int((start + timedelta(days=1)).timestamp()),
                }
            if resource is stripe.Subscription:
                # list subscriptions api does not list canceled subscriptions by default
                # https://stripe.com/docs/api/subscriptions/list
                kwargs["status"] = "all"
            for instance in resource.list(**kwargs).auto_paging_iter():
                row: Dict[str, Any] = {
                    "created": datetime.utcfromtimestamp(instance.created).isoformat()
                }
                if resource is stripe.Event:
                    if not isinstance(instance.data.object, EVENT_DATA_TYPES):
                        continue  # skip events without a defined schema
                    row["data"] = {
                        instance.data.object.object.replace(
                            ".", "_"
                        ): instance.data.object,
                    }
                for key, value in instance.items():
                    if key not in row:
                        row[key] = value
                file_obj.write(ujson.dumps(bigquery_format(row)).encode("UTF-8"))
                file_obj.write(b"\n")
        if table:
            if file_obj.writable():
                file_obj.seek(0)
            warnings.filterwarnings("ignore", module="google.auth._default")
            job = bigquery.Client().load_table_from_file(
                file_obj=file_obj,
                destination=table,
                job_config=bigquery.LoadJobConfig(
                    clustering_fields=["created"],
                    ignore_unknown_values=False,
                    schema=_get_schema(resource),
                    source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
                    time_partitioning=bigquery.TimePartitioning(field="created"),
                    write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                    schema_update_options=[
                        bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION
                    ],
                ),
            )
            try:
                print(f"Waiting for {job.job_id}", file=sys.stderr)
                job.result()
            except Exception as e:
                print(f"{job.job_id} failed: {e}", file=sys.stderr)
                for error in job.errors or ():
                    message = error.get("message")
                    if message and message != getattr(e, "message", None):
                        print(message, file=sys.stderr)
                sys.exit(1)
            else:
                print(f"{job.job_id} succeeded", file=sys.stderr)



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "amiyaguchi@mozilla.com",
    "start_date": datetime.datetime(2020, 2, 17, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "amiyaguchi@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_document_sample", default_args=default_args, schedule_interval="@daily"
) as dag:

    monitoring__document_sample_nonprod__v1 = bigquery_etl_query(
        task_id="monitoring__document_sample_nonprod__v1",
        destination_table="document_sample_nonprod_v1",
        dataset_id="monitoring",
        project_id="moz-fx-data-shared-prod",
        owner="amiyaguchi@mozilla.com",
        email=["amiyaguchi@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        arguments=["--append_table"],
        dag=dag,
    )



#!/usr/bin/env python3

"""
Writes multiple queries to temporary tables and then joins the results.

This is useful for queries that BigQuery deems too complex to run, usually due
to using a large number of subqueries; this pattern allows you to materialize
subsets of columns in multiple different queries so that they stay under the
complexity limit, and then join those results to generate a final wide result.

The query files must be in the same directory and all be prefixed with `part`.
"""

from argparse import ArgumentParser
from multiprocessing.pool import ThreadPool
import os.path

from google.cloud import bigquery

from .util.temp_table import get_temporary_table


def dirpath(string):
    """Path to a dir that must exist."""
    if not os.path.isdir(string):
        raise ValueError()
    return string


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--using",
    default="document_id",
    help="comma separated list of join columns to use when combining results",
)
parser.add_argument(
    "--parallelism",
    default=4,
    type=int,
    help="Maximum number of queries to execute concurrently",
)
parser.add_argument(
    "--dataset_id",
    help="Default dataset, if not specified all tables must be qualified with dataset",
)
parser.add_argument(
    "--project_id", help="Default project, if not specified the sdk will determine one"
)
parser.add_argument(
    "--destination_table",
    required=True,
    help="table where combined results will be written",
)
parser.add_argument(
    "--time_partitioning_field",
    type=lambda f: bigquery.TimePartitioning(field=f),
    help="time partition field on the destination table",
)
parser.add_argument(
    "--clustering_fields",
    type=lambda f: f.split(","),
    help="comma separated list of clustering fields on the destination table",
)
parser.add_argument(
    "--dry_run",
    action="store_true",
    help="Print bytes that would be processed for each part and don't run queries",
)
parser.add_argument(
    "--parameter",
    action="append",
    default=[],
    dest="parameters",
    type=lambda p: bigquery.ScalarQueryParameter(*p.split(":", 2)),
    metavar="NAME:TYPE:VALUE",
    help="query parameter(s) to pass when running parts",
)
parser.add_argument(
    "--priority",
    choices=["BATCH", "INTERACTIVE"],
    default="INTERACTIVE",
    type=str.upper,
    help=(
        "Priority for BigQuery query jobs; BATCH priority will significantly slow "
        "down queries if reserved slots are not enabled for the billing project; "
        "defaults to INTERACTIVE"
    ),
)
parser.add_argument(
    "query_dir",
    type=dirpath,
    help="Path to the query directory that contains part*.sql files",
)
parser.add_argument(
    "--schema_update_option",
    action="append",
    choices=[
        bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION,
        bigquery.SchemaUpdateOption.ALLOW_FIELD_RELAXATION,
        # Airflow passes an empty string when the field addition date doesn't
        # match the run date.
        # See https://github.com/mozilla/telemetry-airflow/blob/
        # e49fa7e6b3f5ec562dd248d257770c2303cf0cba/dags/utils/gcp.py#L515
        "",
    ],
    default=[],
    dest="schema_update_options",
    help="Optional options for updating the schema.",
)


def _sql_table_id(table):
    return f"{table.project}.{table.dataset_id}.{table.table_id}"


def _run_part(client, part, args):
    with open(os.path.join(args.query_dir, part)) as sql_file:
        query = sql_file.read()
    job_config = bigquery.QueryJobConfig(
        destination=get_temporary_table(client),
        default_dataset=args.dataset_id,
        use_legacy_sql=False,
        dry_run=args.dry_run,
        query_parameters=args.parameters,
        priority=args.priority,
        allow_large_results=True,
    )
    job = client.query(query=query, job_config=job_config)
    if job.dry_run:
        print(f"Would process {job.total_bytes_processed:,d} bytes for {part}")
    else:
        job.result()
        print(f"Processed {job.total_bytes_processed:,d} bytes for {part}")
    return part, job


def main():
    """Run multipart query."""
    args = parser.parse_args()
    if (
        args.dataset_id is not None
        and "." not in args.dataset_id
        and args.project_id is not None
    ):
        args.dataset_id = f"{args.project_id}.{args.dataset_id}"
    if "." not in args.destination_table and args.dataset_id is not None:
        args.destination_table = f"{args.dataset_id}.{args.destination_table}"
    client = bigquery.Client(args.project_id)
    with ThreadPool(args.parallelism) as pool:
        parts = pool.starmap(
            _run_part,
            [
                (client, part, args)
                for part in sorted(next(os.walk(args.query_dir))[2])
                if part.startswith("part") and part.endswith(".sql")
            ],
            chunksize=1,
        )
    if not args.dry_run:
        total_bytes = sum(job.total_bytes_processed for _, job in parts)
        query = (
            f"SELECT\n  *\nFROM\n  `{_sql_table_id(parts[0][1].destination)}`"
            + "".join(
                f"\nFULL JOIN\n  `{_sql_table_id(job.destination)}`"
                f"\nUSING\n  ({args.using})"
                for _, job in parts[1:]
            )
        )
        try:
            job = client.query(
                query=query,
                job_config=bigquery.QueryJobConfig(
                    destination=args.destination_table,
                    time_partitioning=args.time_partitioning_field,
                    clustering_fields=args.clustering_fields,
                    write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                    use_legacy_sql=False,
                    priority=args.priority,
                    schema_update_options=args.schema_update_options,
                ),
            )
            job.result()
            print(f"Processed {job.total_bytes_processed:,d} bytes to combine results")
            total_bytes += job.total_bytes_processed
            print(f"Processed {total_bytes:,d} bytes in total")
        finally:
            for _, job in parts:
                client.delete_table(_sql_table_id(job.destination).split("$")[0])
            print(f"Deleted {len(parts)} temporary tables")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Determine cost of previously scheduled bigquery-etl queries."""

from argparse import ArgumentParser
from fnmatch import fnmatchcase
from google.cloud import bigquery
from pathlib import Path

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="monitoring")
parser.add_argument(
    "--destination_table", default="bigquery_etl_scheduled_queries_cost_v1"
)
parser.add_argument("--sql_dir", default="sql/")


def create_query(query_paths, date):
    datasets = ['"' + path.parent.parent.name + '"' for path in query_paths]

    tables = [path.parent.name + "*" for path in query_paths]

    return f"""
      SELECT
        DATE('{date}') AS submission_date,
        creation_time,
        destination_table.dataset_id AS dataset,
        -- remove partition from table name
        REGEXP_REPLACE(destination_table.table_id, r"\$.+", "") AS table,
        total_bytes_processed,
        total_bytes_processed / 1024 / 1024 / 1024 / 1024 * 5 AS cost_usd
      FROM `region-us`.INFORMATION_SCHEMA.JOBS_BY_PROJECT
      WHERE DATE(creation_time) = '{date}'
        AND destination_table.dataset_id IN ({",".join(datasets)})
        AND REGEXP_CONTAINS(destination_table.table_id, "({"|".join(tables)})")
    """


def main():
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    sql_queries = list(Path(args.sql_dir).rglob("query.sql"))
    python_queries = list(Path(args.sql_dir).rglob("query.py"))
    multipart_queries = list(Path(args.sql_dir).rglob("part1.sql"))
    query_paths = sql_queries + python_queries + multipart_queries

    query = create_query(query_paths, args.date)

    partition = args.date.replace("-", "")
    destination_table = f"{args.project}.{args.destination_dataset}.{args.destination_table}${partition}"
    job_config = bigquery.QueryJobConfig(
        destination=destination_table, write_disposition="WRITE_TRUNCATE"
    )
    client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



from pathlib import Path

from bigquery_etl.docs.validate_docs import sql_for_dry_run

TEST_DIR = Path(__file__).parent.parent


class TestValidateDocs:
    def test_sql_for_dry_run_no_dependencies(self):
        file = (
            TEST_DIR
            / "data"
            / "test_docs"
            / "generated_docs"
            / "test_dataset1"
            / "examples"
            / "example1.sql"
        )

        sql = sql_for_dry_run(file, {}, "")
        assert sql == open(file).read()



"""Generating and run baseline_clients_daily queries for Glean apps."""

from argparse import ArgumentParser
from datetime import datetime
from functools import partial
import logging
from multiprocessing.pool import ThreadPool

from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import WriteDisposition, ScalarQueryParameter

from bigquery_etl.util import standard_args  # noqa E402
from bigquery_etl.glean_usage.common import (
    list_baseline_tables,
    render,
    table_names_from_baseline,
    write_sql,
)

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--project_id",
    "--project-id",
    default="moz-fx-data-shar-nonprod-efed",
    help="ID of the project in which to find tables",
)
parser.add_argument(
    "--date",
    required=True,
    type=lambda d: datetime.strptime(d, "%Y-%m-%d").date(),
    help="Date partition to process, in format 2019-01-01",
)
parser.add_argument(
    "--output_dir",
    "--output-dir",
    help="Also write the query text underneath the given sql dir",
)
standard_args.add_parallelism(parser)
standard_args.add_dry_run(parser, debug_log_queries=False)
standard_args.add_log_level(parser)
standard_args.add_priority(parser)
standard_args.add_billing_projects(parser)
standard_args.add_table_filter(parser)


BASELINE_DAILY_TABLE_ID = "baseline_clients_daily_v1"
QUERY_FILENAME = f"{BASELINE_DAILY_TABLE_ID}.sql"
VIEW_FILENAME = f"{BASELINE_DAILY_TABLE_ID[:-3]}.view.sql"


def main():
    """Generate and run queries based on CLI args."""
    args = parser.parse_args()

    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        parser.error(f"argument --log-level: {e}")

    client = bigquery.Client(args.project_id)

    with ThreadPool(args.parallelism) as pool:
        baseline_tables = list_baseline_tables(
            client=client,
            pool=pool,
            project_id=args.project_id,
            only_tables=getattr(args, "only_tables", None),
            table_filter=args.table_filter,
        )
        # Do a first pass with dry_run=True so we don't end up with a partial success;
        # we also write out queries in this pass if so configured.
        pool.map(
            partial(
                run_query,
                client,
                date=args.date,
                dry_run=True,
                output_dir=args.output_dir,
            ),
            baseline_tables,
        )
        logging.info(
            f"Dry runs successful for {len(baseline_tables)}"
            " baseline_clients_daily table(s)"
        )
        # Now, actually run the queries.
        if not args.dry_run:
            pool.map(
                partial(run_query, client, date=args.date, dry_run=False),
                baseline_tables,
            )


def run_query(client, baseline_table, date, dry_run, output_dir=None):
    """Process a single table, potentially also writing out the generated queries."""
    tables = table_names_from_baseline(baseline_table)

    daily_table = tables["daily_table"]
    daily_view = tables["daily_view"]
    render_kwargs = dict(header="-- Generated via bigquery_etl.glean_usage\n")
    render_kwargs.update(tables)
    job_kwargs = dict(use_legacy_sql=False, dry_run=dry_run)

    sql = render(QUERY_FILENAME, **render_kwargs)
    init_sql = render(QUERY_FILENAME, init=True, **render_kwargs)
    view_sql = render(VIEW_FILENAME, **render_kwargs)
    if output_dir:
        write_sql(output_dir, daily_table, "query.sql", sql)
        write_sql(output_dir, daily_table, "init.sql", init_sql)
        write_sql(output_dir, daily_view, "view.sql", view_sql)

    try:
        client.get_table(daily_table)
    except NotFound:
        if dry_run:
            logging.info(f"Table does not yet exist: {daily_table}")
        else:
            logging.info(f"Creating table: {daily_table}")
        sql = init_sql
    else:
        # Table exists, so we will run the incremental query.
        job_kwargs.update(
            destination=f"{daily_table}${date.strftime('%Y%m%d')}",
            write_disposition=WriteDisposition.WRITE_TRUNCATE,
            query_parameters=[ScalarQueryParameter("submission_date", "DATE", date)],
        )
        if not dry_run:
            logging.info(f"Running query for: {daily_table}")
    job_config = bigquery.QueryJobConfig(**job_kwargs)
    job = client.query(sql, job_config)
    if not dry_run:
        job.result()
        logging.info(f"Recreating view {daily_view}")
        client.query(view_sql, bigquery.QueryJobConfig(use_legacy_sql=False)).result()


if __name__ == "__main__":
    main()



from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.metadata.validate_metadata import validate_public_data


class TestValidateMetadata(object):
    def test_is_valid_public_data(self):
        metadata_not_public = Metadata("No public data", "No public data", {}, {})
        assert validate_public_data(metadata_not_public, "test/path/metadata.yaml")

        metadata_valid_public = Metadata(
            "Public json data",
            "Public json data",
            [],
            {"public_json": True, "review_bugs": [123456]},
            {},
        )
        assert validate_public_data(metadata_valid_public, "test/path/metadata.yaml")

        metadata_valid_public = Metadata(
            "Public BigQuery data",
            "Public BigQuery data",
            [],
            {"public_bigquery": True, "review_bugs": [123456]},
            {},
        )
        assert validate_public_data(metadata_valid_public, "test/path/metadata.yaml")

        metadata_invalid_public = Metadata(
            "Public BigQuery data",
            "Public BigQuery data",
            [],
            {"public_bigquery": True},
            {},
        )
        assert (
            validate_public_data(metadata_invalid_public, "test/path/metadata.yaml")
            is False
        )



import os

from bigquery_etl.dryrun import DryRun


class TestDryRun:
    def test_dry_run_sql_file(self, tmp_path):
        query_file = tmp_path / "query.sql"
        query_file.write_text("SELECT 123")

        dryrun = DryRun(str(query_file))
        response = dryrun.dry_run_result
        assert response["valid"]

    def test_dry_run_invalid_sql_file(self, tmp_path):
        query_file = tmp_path / "query.sql"
        query_file.write_text("SELECT INVALID 123")

        dryrun = DryRun(str(query_file))
        response = dryrun.dry_run_result
        assert response["valid"] is False

    def test_sql_file_valid(self, tmp_path):
        query_file = tmp_path / "query.sql"
        query_file.write_text("SELECT 123")

        dryrun = DryRun(str(query_file))
        assert dryrun.is_valid()

    def test_sql_file_invalid(self, tmp_path):
        query_file = tmp_path / "query.sql"
        query_file.write_text("SELECT INVALID 123")

        dryrun = DryRun(str(query_file))
        assert dryrun.is_valid() is False

    def test_get_referenced_tables_empty(self, tmp_path):
        query_file = tmp_path / "query.sql"
        query_file.write_text("SELECT 123")

        dryrun = DryRun(str(query_file))
        assert dryrun.get_referenced_tables() == []

    def test_get_referenced_tables(self, tmp_path):
        os.makedirs(tmp_path / "telmetry_derived")
        query_file = tmp_path / "telmetry_derived" / "query.sql"
        query_file.write_text(
            "SELECT * FROM telemetry_derived.clients_daily_v6 "
            "WHERE submission_date = '2020-01-01'"
        )
        dryrun = DryRun(str(query_file))
        response = dryrun.get_referenced_tables()

        assert len(response) == 1
        assert response[0]["datasetId"] == "telemetry_derived"
        assert response[0]["tableId"] == "clients_daily_v6"



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "ascholtz@mozilla.com",
    "start_date": datetime.datetime(2018, 10, 30, 0, 0),
    "end_date": None,
    "email": ["ascholtz@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_monitoring", default_args=default_args, schedule_interval="0 2 * * *"
) as dag:

    monitoring__bigquery_etl_scheduled_queries_cost__v1 = gke_command(
        task_id="monitoring__bigquery_etl_scheduled_queries_cost__v1",
        command=[
            "python",
            "sql/moz-fx-data-shared-prod/monitoring/bigquery_etl_scheduled_queries_cost_v1/query.py",
        ]
        + ["--date", "{{ ds }}"],
        docker_image="mozilla/bigquery-etl:latest",
        owner="ascholtz@mozilla.com",
        email=["ascholtz@mozilla.com"],
    )

    monitoring__bigquery_etl_scheduled_query_usage__v1 = gke_command(
        task_id="monitoring__bigquery_etl_scheduled_query_usage__v1",
        command=[
            "python",
            "sql/moz-fx-data-shared-prod/monitoring/bigquery_etl_scheduled_query_usage_v1/query.py",
        ]
        + ["--date", "{{ ds }}"],
        docker_image="mozilla/bigquery-etl:latest",
        owner="ascholtz@mozilla.com",
        email=["ascholtz@mozilla.com"],
    )

    monitoring__column_size__v1 = gke_command(
        task_id="monitoring__column_size__v1",
        command=[
            "python",
            "sql/moz-fx-data-shared-prod/monitoring/column_size_v1/query.py",
        ]
        + ["--date", "{{ ds }}"],
        docker_image="mozilla/bigquery-etl:latest",
        owner="ascholtz@mozilla.com",
        email=["ascholtz@mozilla.com"],
    )

    monitoring__schema_error_counts__v2 = bigquery_etl_query(
        task_id="monitoring__schema_error_counts__v2",
        destination_table="schema_error_counts_v2",
        dataset_id="monitoring",
        project_id="moz-fx-data-shared-prod",
        owner="amiyaguchi@mozilla.com",
        email=["amiyaguchi@mozilla.com", "ascholtz@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    monitoring__stable_table_sizes__v1 = gke_command(
        task_id="monitoring__stable_table_sizes__v1",
        command=[
            "python",
            "sql/moz-fx-data-shared-prod/monitoring/stable_table_sizes_v1/query.py",
        ]
        + ["--date", "{{ ds }}"],
        docker_image="mozilla/bigquery-etl:latest",
        owner="ascholtz@mozilla.com",
        email=["ascholtz@mozilla.com"],
    )

    monitoring__structured_distinct_docids__v1 = gke_command(
        task_id="monitoring__structured_distinct_docids__v1",
        command=[
            "python",
            "sql/moz-fx-data-shared-prod/monitoring/structured_distinct_docids_v1/query.py",
        ]
        + ["--date", "{{ ds }}"],
        docker_image="mozilla/bigquery-etl:latest",
        owner="bewu@mozilla.com",
        email=["ascholtz@mozilla.com", "bewu@mozilla.com"],
    )

    monitoring__structured_missing_columns__v1 = gke_command(
        task_id="monitoring__structured_missing_columns__v1",
        command=[
            "python",
            "sql/moz-fx-data-shared-prod/monitoring/structured_missing_columns_v1/query.py",
        ]
        + ["--date", "{{ ds }}"],
        docker_image="mozilla/bigquery-etl:latest",
        owner="amiyaguchi@mozilla.com",
        email=["amiyaguchi@mozilla.com", "ascholtz@mozilla.com"],
    )

    monitoring__telemetry_distinct_docids__v1 = bigquery_etl_query(
        task_id="monitoring__telemetry_distinct_docids__v1",
        destination_table="telemetry_distinct_docids_v1",
        dataset_id="monitoring",
        project_id="moz-fx-data-shared-prod",
        owner="bewu@mozilla.com",
        email=["ascholtz@mozilla.com", "bewu@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    monitoring__telemetry_missing_columns__v3 = bigquery_etl_query(
        task_id="monitoring__telemetry_missing_columns__v3",
        destination_table="telemetry_missing_columns_v3",
        dataset_id="monitoring",
        project_id="moz-fx-data-shared-prod",
        owner="amiyaguchi@mozilla.com",
        email=["amiyaguchi@mozilla.com", "ascholtz@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    wait_for_copy_deduplicate_main_ping = ExternalTaskSensor(
        task_id="wait_for_copy_deduplicate_main_ping",
        external_dag_id="copy_deduplicate",
        external_task_id="copy_deduplicate_main_ping",
        execution_delta=datetime.timedelta(seconds=3600),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    monitoring__column_size__v1.set_upstream(wait_for_copy_deduplicate_main_ping)

    wait_for_copy_deduplicate_all = ExternalTaskSensor(
        task_id="wait_for_copy_deduplicate_all",
        external_dag_id="copy_deduplicate",
        external_task_id="copy_deduplicate_all",
        execution_delta=datetime.timedelta(seconds=3600),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    monitoring__stable_table_sizes__v1.set_upstream(wait_for_copy_deduplicate_all)
    monitoring__stable_table_sizes__v1.set_upstream(wait_for_copy_deduplicate_main_ping)

    monitoring__structured_distinct_docids__v1.set_upstream(
        wait_for_copy_deduplicate_all
    )
    monitoring__structured_distinct_docids__v1.set_upstream(
        wait_for_copy_deduplicate_main_ping
    )

    monitoring__structured_missing_columns__v1.set_upstream(
        wait_for_copy_deduplicate_all
    )

    monitoring__telemetry_distinct_docids__v1.set_upstream(
        wait_for_copy_deduplicate_all
    )
    monitoring__telemetry_distinct_docids__v1.set_upstream(
        wait_for_copy_deduplicate_main_ping
    )



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "dthorn@mozilla.com",
    "start_date": datetime.datetime(2020, 10, 5, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "dthorn@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=300),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG("bqetl_stripe", default_args=default_args, schedule_interval="@daily") as dag:

    stripe_derived__customers__v1 = bigquery_etl_query(
        task_id="stripe_derived__customers__v1",
        destination_table="customers_v1",
        dataset_id="stripe_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    stripe_derived__plans__v1 = bigquery_etl_query(
        task_id="stripe_derived__plans__v1",
        destination_table="plans_v1",
        dataset_id="stripe_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    stripe_derived__products__v1 = bigquery_etl_query(
        task_id="stripe_derived__products__v1",
        destination_table="products_v1",
        dataset_id="stripe_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    stripe_derived__subscriptions__v1 = bigquery_etl_query(
        task_id="stripe_derived__subscriptions__v1",
        destination_table="subscriptions_v1",
        dataset_id="stripe_derived",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    stripe_external__charges__v1 = bigquery_etl_query(
        task_id="stripe_external__charges__v1",
        destination_table="charges_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__credit_notes__v1 = bigquery_etl_query(
        task_id="stripe_external__credit_notes__v1",
        destination_table="credit_notes_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__customers__v1 = bigquery_etl_query(
        task_id="stripe_external__customers__v1",
        destination_table="customers_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__disputes__v1 = bigquery_etl_query(
        task_id="stripe_external__disputes__v1",
        destination_table="disputes_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__invoices__v1 = bigquery_etl_query(
        task_id="stripe_external__invoices__v1",
        destination_table="invoices_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__payment_intents__v1 = bigquery_etl_query(
        task_id="stripe_external__payment_intents__v1",
        destination_table="payment_intents_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__payouts__v1 = bigquery_etl_query(
        task_id="stripe_external__payouts__v1",
        destination_table="payouts_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__plans__v1 = bigquery_etl_query(
        task_id="stripe_external__plans__v1",
        destination_table="plans_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__prices__v1 = bigquery_etl_query(
        task_id="stripe_external__prices__v1",
        destination_table="prices_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__products__v1 = bigquery_etl_query(
        task_id="stripe_external__products__v1",
        destination_table="products_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__setup_intents__v1 = bigquery_etl_query(
        task_id="stripe_external__setup_intents__v1",
        destination_table="setup_intents_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_external__subscriptions__v1 = bigquery_etl_query(
        task_id="stripe_external__subscriptions__v1",
        destination_table="subscriptions_v1",
        dataset_id="stripe_external",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        parameters=["date:DATE:{{ds}}"],
        dag=dag,
    )

    stripe_derived__customers__v1.set_upstream(stripe_external__customers__v1)

    stripe_derived__plans__v1.set_upstream(stripe_external__plans__v1)

    stripe_derived__products__v1.set_upstream(stripe_external__products__v1)

    stripe_derived__subscriptions__v1.set_upstream(stripe_external__subscriptions__v1)

    wait_for_stripe_stripe_import_events = ExternalTaskSensor(
        task_id="wait_for_stripe_stripe_import_events",
        external_dag_id="stripe",
        external_task_id="stripe_import_events",
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
        dag=dag,
    )

    stripe_external__charges__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__credit_notes__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__customers__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__disputes__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__invoices__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__payment_intents__v1.set_upstream(
        wait_for_stripe_stripe_import_events
    )

    stripe_external__payouts__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__plans__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__prices__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__products__v1.set_upstream(wait_for_stripe_stripe_import_events)

    stripe_external__setup_intents__v1.set_upstream(
        wait_for_stripe_stripe_import_events
    )

    stripe_external__subscriptions__v1.set_upstream(
        wait_for_stripe_stripe_import_events
    )



"""bigquery-etl CLI dryrun command."""

import os
import sys
from multiprocessing.pool import ThreadPool
from pathlib import Path

import click
from google.cloud import bigquery

from ..cli.utils import is_authenticated
from ..dryrun import SKIP, DryRun


@click.command(
    help="Dry run SQL.",
)
@click.argument(
    "path",
    default="sql/",
    type=click.Path(file_okay=True),
)
@click.option(
    "--use_cloud_function",
    "--use-cloud-function",
    help=(
        "Use the Cloud Function for dry running SQL, if set to `True`. "
        "The Cloud Function can only access tables in shared-prod. "
        "If set to `False`, use active GCP credentials for the dry run."
    ),
    type=bool,
    default=True,
)
@click.option(
    "--project",
    help="GCP project to perform dry run in when --use_cloud_function=False",
    default="moz-fx-data-shared-prod",
)
def dryrun(path, use_cloud_function, project):
    """Perform a dry run."""
    if os.path.isdir(path) and os.path.exists(path):
        sql_files = [f for f in Path(path).rglob("*.sql") if str(f) not in SKIP]
    elif os.path.isfile(path) and os.path.exists(path):
        sql_files = [path]
    else:
        click.echo(f"Invalid path {path}", err=True)
        sys.exit(1)

    if use_cloud_function:

        def cloud_function_dryrun(sqlfile):
            """Dry run SQL files."""
            return DryRun(sqlfile).is_valid()

        sql_file_valid = cloud_function_dryrun
    else:
        if not is_authenticated():
            click.echo("Not authenticated to GCP. Run `gcloud auth login` to login.")
            sys.exit(1)

        client = bigquery.Client()

        def gcp_dryrun(sqlfile):
            """Dry run the SQL file."""
            dataset = Path(sqlfile).parent.parent.name
            job_config = bigquery.QueryJobConfig(
                dry_run=True,
                use_query_cache=False,
                default_dataset=f"{project}.{dataset}",
                query_parameters=[
                    bigquery.ScalarQueryParameter(
                        "submission_date", "DATE", "2019-01-01"
                    )
                ],
            )

            with open(sqlfile) as query_stream:
                query = query_stream.read()

                try:
                    client.query(query, job_config=job_config)
                    click.echo(f"{sqlfile:59} OK")
                    return True
                except Exception as e:
                    click.echo(f"{sqlfile:59} ERROR: {e}")
                    return False

        sql_file_valid = gcp_dryrun

    with ThreadPool(8) as p:
        result = p.map(sql_file_valid, sql_files, chunksize=1)
    if not all(result):
        sys.exit(1)



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "ssuh@mozilla.com",
    "start_date": datetime.datetime(2020, 12, 14, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "ssuh@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_search_dashboard", default_args=default_args, schedule_interval="0 4 * * *"
) as dag:

    search_derived__desktop_search_aggregates_by_userstate__v1 = bigquery_etl_query(
        task_id="search_derived__desktop_search_aggregates_by_userstate__v1",
        destination_table="desktop_search_aggregates_by_userstate_v1",
        dataset_id="search_derived",
        project_id="moz-fx-data-shared-prod",
        owner="xluo@mozilla.com",
        email=["ssuh@mozilla.com", "telemetry-alerts@mozilla.com", "xluo@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    search_derived__desktop_search_aggregates_for_searchreport__v1 = bigquery_etl_query(
        task_id="search_derived__desktop_search_aggregates_for_searchreport__v1",
        destination_table="desktop_search_aggregates_for_searchreport_v1",
        dataset_id="search_derived",
        project_id="moz-fx-data-shared-prod",
        owner="xluo@mozilla.com",
        email=["ssuh@mozilla.com", "telemetry-alerts@mozilla.com", "xluo@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    search_derived__mobile_search_aggregates_for_searchreport__v1 = bigquery_etl_query(
        task_id="search_derived__mobile_search_aggregates_for_searchreport__v1",
        destination_table="mobile_search_aggregates_for_searchreport_v1",
        dataset_id="search_derived",
        project_id="moz-fx-data-shared-prod",
        owner="mmccorquodale@mozilla.com",
        email=[
            "mmccorquodale@mozilla.com",
            "ssuh@mozilla.com",
            "telemetry-alerts@mozilla.com",
            "xluo@mozilla.com",
        ],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    wait_for_telemetry_derived__clients_last_seen__v1 = ExternalTaskSensor(
        task_id="wait_for_telemetry_derived__clients_last_seen__v1",
        external_dag_id="bqetl_main_summary",
        external_task_id="telemetry_derived__clients_last_seen__v1",
        execution_delta=datetime.timedelta(seconds=7200),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    search_derived__desktop_search_aggregates_by_userstate__v1.set_upstream(
        wait_for_telemetry_derived__clients_last_seen__v1
    )

    wait_for_search_derived__search_aggregates__v8 = ExternalTaskSensor(
        task_id="wait_for_search_derived__search_aggregates__v8",
        external_dag_id="bqetl_search",
        external_task_id="search_derived__search_aggregates__v8",
        execution_delta=datetime.timedelta(seconds=3600),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    search_derived__desktop_search_aggregates_for_searchreport__v1.set_upstream(
        wait_for_search_derived__search_aggregates__v8
    )

    wait_for_search_derived__mobile_search_clients_daily__v1 = ExternalTaskSensor(
        task_id="wait_for_search_derived__mobile_search_clients_daily__v1",
        external_dag_id="bqetl_mobile_search",
        external_task_id="search_derived__mobile_search_clients_daily__v1",
        execution_delta=datetime.timedelta(seconds=7200),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    search_derived__mobile_search_aggregates_for_searchreport__v1.set_upstream(
        wait_for_search_derived__mobile_search_clients_daily__v1
    )



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "bewu@mozilla.com",
    "start_date": datetime.datetime(2020, 11, 30, 0, 0),
    "end_date": None,
    "email": ["bewu@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_marketing_fetch", default_args=default_args, schedule_interval="0 1 * * 1"
) as dag:

    fetch__spend_alignment_by_campaign__v1 = bigquery_etl_query(
        task_id="fetch__spend_alignment_by_campaign__v1",
        destination_table="spend_alignment_by_campaign_v1",
        dataset_id="fetch",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )

    fetch__spend_alignment_by_month__v1 = bigquery_etl_query(
        task_id="fetch__spend_alignment_by_month__v1",
        destination_table="spend_alignment_by_month_v1",
        dataset_id="fetch",
        project_id="moz-fx-data-marketing-prod",
        owner="bewu@mozilla.com",
        email=["bewu@mozilla.com"],
        date_partition_parameter=None,
        depends_on_past=False,
        dag=dag,
    )



"""Validate metadata files."""

from argparse import ArgumentParser
import logging
import os
import sys

from .parse_metadata import Metadata
from ..util import standard_args
from ..util.common import project_dirs

parser = ArgumentParser(description=__doc__)

parser.add_argument("--target", help="File or directory containing metadata files")
standard_args.add_log_level(parser)


def validate_public_data(metadata, path):
    """Check if the metadata for public data queries is valid."""
    is_valid = True

    if metadata.is_public_bigquery() or metadata.is_public_json():
        if not metadata.review_bugs():
            logging.error(f"Missing review bug for public data: {path}")
            is_valid = False

    return is_valid


def validate(target):
    """Validate metadata files."""
    failed = False

    if os.path.isdir(target):
        for root, dirs, files in os.walk(target):
            for file in files:
                if Metadata.is_metadata_file(file):
                    path = os.path.join(root, *dirs, file)
                    metadata = Metadata.from_file(path)

                    if not validate_public_data(metadata, path):
                        failed = True

                    # todo more validation
                    # e.g. https://github.com/mozilla/bigquery-etl/issues/924
    else:
        logging.error(f"Invalid target: {target}, target must be a directory.")
        sys.exit(1)

    if failed:
        sys.exit(1)


def main():
    """Validate all metadata.yaml files in the provided target directory."""
    args = parser.parse_args()

    # set log level
    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        parser.error(f"argument --log-level: {e}")

    if args.target:
        validate(args.target)
    else:
        for project in project_dirs():
            validate(project)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Experiment search aggregates live view query generator."""
import argparse
import datetime
import json
import sys
import textwrap
from time import sleep

p = argparse.ArgumentParser()
p.add_argument(
    "--submission-date",
    type=str,
    help="Cut-off date for using pre-computed vs live tables in view",
    required=True,
)
p.add_argument(
    "--json-output",
    action="store_true",
    help="Output the result wrapped in json parseable as an XCOM",
)
p.add_argument(
    "--wait-seconds",
    type=int,
    default=0,
    help="Add a delay before executing the script to allow time for the xcom sidecar to complete startup",
)


def generate_sql(opts):
    """Create a BigQuery SQL query for the experiment search aggregates view with new date filled in.
       Unfortunately, BigQuery does not allow parameters in view definitions, so this script is a very thin
       wrapper over the nearly complete sql query to fill in the cutoff date for using the pre-aggregated vs
       live table for the view"""
    query = textwrap.dedent(
        """
        CREATE OR REPLACE VIEW
          `moz-fx-data-shared-prod.telemetry.experiment_search_aggregates_live_v1`
        AS
        WITH all_searches AS (
          SELECT
            *
          FROM
            `moz-fx-data-shared-prod.telemetry_derived.experiment_search_aggregates_v1`
          WHERE
            date(window_start) <= DATE '{submission_date}'
          UNION ALL
          SELECT
            * EXCEPT (submission_date, dataset_id)
          FROM
            `moz-fx-data-shared-prod.telemetry_derived.experiment_search_aggregates_base`
          WHERE
            submission_date > DATE '{submission_date}'
            AND dataset_id = 'telemetry_live'
        )
        SELECT
          *,
          SUM(search_count) OVER previous_rows_window AS cumulative_search_count,
          SUM(search_with_ads_count) OVER previous_rows_window AS cumulative_search_with_ads_count,
          SUM(ad_clicks_count) OVER previous_rows_window AS cumulative_ad_clicks_count
        FROM
          all_searches
        WINDOW previous_rows_window AS (
          PARTITION BY
            experiment,
            branch
          ORDER BY
            window_start
          ROWS BETWEEN
            UNBOUNDED PRECEDING
            AND CURRENT ROW
        )
        """.format(
            **opts
        )
    )
    if opts["json_output"]:
        return json.dumps(query)
    else:
        return query


def main(argv, out=print):
    """Print experiment search aggregates view sql to stdout."""
    opts = vars(p.parse_args(argv[1:]))
    sleep(opts["wait_seconds"])
    out(generate_sql(opts))


if __name__ == "__main__":
    main(sys.argv)



"""bigquery-etl CLI."""

import warnings

import click

from .._version import __version__
from ..cli.dag import dag
from ..cli.dryrun import dryrun
from ..cli.format import format
from ..cli.query import query
from ..cli.routine import mozfun, routine
from ..cli.view import view
from ..glam.cli import glam
from ..stripe import stripe_


def cli(prog_name=None):
    """Create the bigquery-etl CLI."""
    commands = {
        "query": query,
        "dag": dag,
        "dryrun": dryrun,
        "format": format,
        "routine": routine,
        "mozfun": mozfun,
        "stripe": stripe_,
        "glam": glam,
        "view": view,
    }

    @click.group(commands=commands)
    @click.version_option(version=__version__)
    def group():
        """CLI tools for working with bigquery-etl."""
        pass

    warnings.filterwarnings(
        "ignore", "Your application has authenticated using end user credentials"
    )

    group(prog_name=prog_name)


if __name__ == "__main__":
    cli()



from setuptools import setup, find_namespace_packages

def get_version():
    version = {}
    with open('bigquery_etl/_version.py') as fp:
        exec(fp.read(), version)

    return version['__version__']


setup(
    name="mozilla-bigquery-etl",
    version=get_version(),
    author="Mozilla Corporation",
    author_email="fx-data-dev@mozilla.org",
    description="Tooling for building derived datasets in BigQuery",
    url="https://github.com/mozilla/bigquery-etl",
    packages=find_namespace_packages(include=["bigquery_etl.*", "bigquery_etl"]),
    package_data={'bigquery_etl': ['query_scheduling/templates/*.j2']},
    include_package_data=True,
    install_requires=[
        "gcloud",
        "google-cloud-bigquery",
        "google-cloud-storage",
        "Jinja2",
        "pytest-black",
        "pytest-docstyle",
        "pytest-flake8",
        "pytest-mypy",
        "pytest",
        "PyYAML",
        "smart_open",
        "sqlparse",
        "mozilla_schema_generator",
        "GitPython",
        "cattrs",
        "attrs",
        "typing",
        "click",
        "pandas",
        "ujson",
        "stripe",
    ],
    long_description="Tooling for building derived datasets in BigQuery",
    long_description_content_type="text/markdown",
    python_requires="==3.8.*",
    entry_points="""
        [console_scripts]
        bqetl=bigquery_etl.cli:cli
    """,
)
